from blabpy.vihi.paths import get_vihi_path, get_subject_files_path, get_lena_path


def test_get_vihi_path():
    vihi_path = get_vihi_path()
    assert vihi_path.name == 'VIHI'
    assert vihi_path.exists()


def test_get_subject_files_path():
    subject_files_path = get_subject_files_path()
    assert subject_files_path.name == 'SubjectFiles'
    assert subject_files_path.exists()


def test_get_lena_path():
    lena_path = get_lena_path()
    assert lena_path.name == 'LENA'
    assert lena_path.exists()



# Initially, an exact copy of https://github.com/SeedlingsBabylab/parse_clan2/blob/master/parse_clan2.py
import csv
import logging
import re
import os
import collections
from pathlib import Path


class Parser:

    def __init__(self, input_path, output=None):

        self.input_file = input_path
        # output == None means we're being called from parse_clan2 main
        if output is None:
            self.output_file = input_path.replace(".cha", "_processed.csv")
            self.error_file = input_path.replace(".cha", "_errors.txt")
        # otherwise, Parser object is being constructed from an external source
        else:
            processed_path = os.path.split(input_path)[1].replace(".cha", "_processed.csv")
            error_path = os.path.split(input_path)[1].replace(".cha", "_errors.txt")
            # self.output_file = os.path.join(output, name)
            logging.info(os.path.join(output, processed_path))
            self.output_file = os.path.join(output, processed_path)
            self.error_file = os.path.join(output, error_path)

        # correct regex for annotations
        re1 = '((?:[a-z][a-z0-9_+]*))'  # the word
        re2 = '(\\s+)'	                # whitespace
        re3 = '(&=)'	                # &=
        re4 = '(.)'	                    # utterance_type
        re5 = '(_+)'	                # _
        re6 = '(.)'	                    # object_present
        re7 = '(_+)'	                # _
        re8 = '((?:[a-z][a-z0-9]*))'    # speaker
        re81 = '(_+)?'
        re82 = '(0x[a-z0-9]{6})?'       # annotid

        # incorrect regexes (for typos and formatting issues)
        self.entry_regx = re.compile(re1+re2+re3+re4+re5+re6+re7+re8+re81+re82, re.IGNORECASE | re.DOTALL)
        self.old_entry_regx = re.compile(re1+re2+'(&)'+re4+'(\\|)'+re6+'(\\|)'+re8+re81+re82, re.IGNORECASE | re.DOTALL)
        self.interval_regx = re.compile("(\025\d+_\d+)")

        self.joined_num_regx = re.compile("(_[yn]_[a-z0-9]{3}\d+)", re.IGNORECASE | re.DOTALL)
        self.joined_entry_wrdcount = re.compile("(_[yn]_[a-z0-9]{3}&=w)", re.IGNORECASE | re.DOTALL)
        # for new format:
        # self.joined_entry_wrdcount = re.compile("([a-f0-9]{6}&=w)", re.IGNORECASE | re.DOTALL)
        self.just_ampersand_regx = re.compile(re1+re2+'(&)'+'([qdiursn])'+re5+re6+re7+re8+re81+re82, re.IGNORECASE | re.DOTALL)

        # someword &=d-y-MOT
        self.dash_not_underscore_all = re.compile(re1+re2+re3+re4+'(-+)'+re6+'(-+)'+re8+re81+re82, re.IGNORECASE | re.DOTALL)
        # someword &=d-y_MOT
        self.dash_not_underscore_first = re.compile(re1+re2+re3+re4+'(-+)'+re6+re7+re8+re81+re82, re.IGNORECASE | re.DOTALL)
        # someword &=d_y-MOT
        self.dash_not_underscore_second = re.compile(re1+re2+re3+re4+re5+re6+'(-+)'+re8+re81+re82, re.IGNORECASE | re.DOTALL)

        re9 = '((?:[a-z][a-z]+))'  # Word 1
        re10 = '(\\s+)'            # White Space 1
        re11 = '(0)'               # Any Single Character 1
        re12 = '(\\s+)'            # White Space 2
        re13 = '(.)'               # Any Single Character 2

        # tummy 0 .
        self.missing_code_just_word = re.compile('(\\s+)'+re9+re10+re11+re12+re13,re.IGNORECASE|re.DOTALL)
        # tummy 0.
        self.missing_code_just_word_zero_period = re.compile('(\\s+)'+re9+re10+'(0.)'+re12+re13,re.IGNORECASE|re.DOTALL)
        # tummy &=w
        self.missing_code_just_word_andcount = re.compile('(\\s+)'+re9+re10+'(&=w)',re.IGNORECASE|re.DOTALL)

        self.one_missing_code_first = re.compile(re1+re2+re3+re5+'([yn])'+re7+re8+re81+re82, re.IGNORECASE|re.DOTALL)
        self.one_missing_code_second = re.compile(re1+re2+re3+re4+re5+re7+re8+re81+re82, re.IGNORECASE|re.DOTALL)
        self.one_missing_code_third = re.compile(re1+re2+re3+'([qdiursn])'+re5+'([yn])'+re7+'(\\s+)',
                                                 re.IGNORECASE|re.DOTALL)  # TODO
        self.one_missing_code_third_joined_count = re.compile(re1+re2+re3+'([qdiursn])'+re5+'([yn])'+re7+'(&=w)',
                                                              re.IGNORECASE|re.DOTALL)  #TODO

        self.missing_underscore_first = re.compile(re3+'([qdiursn])'+'([yn])'+re5+re8+re81+re82, re.IGNORECASE|re.DOTALL)
        self.missing_underscore_second = re.compile(re3+'([qdiursn])'+re5+'([yn])'+re8+re81+re82, re.IGNORECASE|re.DOTALL)

        # self.scrub_regx = re.compile()

        self.skipping = False
        self.begin_skip_start = None
        self.words = []
        self.comments = []          # includes all the comments
        self.curr_personal_block = None
        self.personal_info_groups = []
        self.plain_comments = []    # not including subregion/silence comments

        # list of tuples representing problems that were found.
        # represented as:
        #
        #       ("Description of problem: ...", interval, problem_entry)
        self.problems = []

        # check_intervals() returns false if there was a problem,
        # so we'll break from parsing and not go any further
        if not self.check_intervals():
            return

        self.phonetic_transcriptions = None

        self.parse()
        self.filter_comments()
        self.export()

    def parse(self):
        last_line = ""
        multi_line = ""

        prev_interval = [None, None]
        curr_interval = [None, None]

        inside_personal_block = False

        personal_info_com_start = None
        personal_info_com_end = None

        with open(self.input_file, "rU") as input:
            for index, line in enumerate(input):

                if line.startswith("%com:") and ("|" not in line):
                    if "begin skip" in line:
                        logging.info("Begin skip found in line# " + str(index))  #+ "\n"
                        self.skipping = True
                        self.begin_skip_start = index
                        continue
                    if "end skip" in line:
                        logging.info("End skip found in line# " + str(index) + "\n")
                        self.skipping = False
                        self.begin_skip_start = None
                        continue
                    if "personal" in line or "private" in line:
                        inside_personal_block = True
                        self.check_personal_info_comment(line, index, curr_interval)
                    if "end personal" in line:
                        inside_personal_block = False
                        self.personal_info_groups.append(self.curr_personal_block)

                    # get rid of quotation marks, %com's and newlines
                    comment = line.replace("%com:\t", "")\
                                  .replace("\"", "")\
                                  .replace("\n", "")

                    self.comments.append((comment, curr_interval[0], curr_interval[1]))

                if (line.startswith("%xcom:")) and ("|" not in line):
                    if "begin skip" in line:
                        logging.info("Begin skip starts at line# " + str(index))  #+ "\n"
                        self.skipping = True
                        self.begin_skip_start = index
                        continue
                    if "end skip" in line:
                        logging.info("End skip found in line# " + str(index) + "\n")
                        self.skipping = False
                        self.begin_skip_start = None
                        continue
                    if "personal" in line or "private" in line:
                        inside_personal_block = True
                    # get rid of quotation marks, %xcom's and newlines
                    comment = line.replace("%xcom:\t", "")\
                                  .replace("\"", "")\
                                  .replace("\n", "")

                    self.comments.append((comment, curr_interval[0], curr_interval[1]))

                #
                # if self.skipping:
                #     print "skipping line: " + line
                #     continue

                if line.startswith("*"):

                    # reset multi_line
                    multi_line = ""
                    interval_reg_result = self.interval_regx.search(line)

                    if interval_reg_result is None:
                        logging.info("interval regx returned none. clan line: " + str(index))
                        last_line = line
                        multi_line += line
                        continue
                    # rearrange previous and current intervals
                    prev_interval[0] = curr_interval[0]
                    prev_interval[1] = curr_interval[1]

                    # set the new curr_interval
                    interval_str = interval_reg_result.group().replace("\025", "")
                    interval = interval_str.split("_")
                    curr_interval[0] = int(interval[0])
                    curr_interval[1] = int(interval[1])

                    # find correctly formatted entries
                    entries = self.entry_regx.findall(line)

                    # check for all the possible malformed entry types
                    old_entries = self.old_entry_regx.findall(line)
                    joined_num = self.joined_num_regx.findall(line)
                    joined_entry_wrdcount = self.joined_entry_wrdcount.findall(line)
                    just_ampersand = self.just_ampersand_regx.findall(line)
                    dash_not_underscore_all = self.dash_not_underscore_all.findall(line)
                    dash_not_underscore_first = self.dash_not_underscore_first.findall(line)
                    dash_not_underscore_second = self.dash_not_underscore_second.findall(line)
                    missing_code_just_word = self.missing_code_just_word.findall(line)
                    missing_code_just_word_zero_period = self.missing_code_just_word_zero_period.findall(line)
                    missing_code_just_word_andcount = self.missing_code_just_word_andcount.findall(line)
                    one_missing_code_first = self.one_missing_code_first.findall(line)
                    one_missing_code_second = self.one_missing_code_second.findall(line)
                    one_missing_code_third = self.one_missing_code_third.findall(line)
                    one_missing_code_third_joined_count = self.one_missing_code_third_joined_count.findall(line)
                    missing_underscore_first = self.missing_underscore_first.findall(line)
                    missing_underscore_second = self.missing_underscore_second.findall(line)

                    if line.startswith("*SCR:"):
                        self.parse_scrub_tier(line, index, interval)

                    # e.g. - someword &=d_y_MOT0 .
                    if joined_num:
                        temp = [""] * len(joined_num)

                        for entry_index, entry in enumerate(joined_num):
                            for group in entry:
                                temp[entry_index] += group
                        self.problems.append(("Joined Number and Speaker Code",
                                              "line#: {}".format(index-1),
                                              interval,
                                              temp))

                    # e.g. - someword &d|y|MOT
                    if old_entries:
                        temp = [""] * len(old_entries)

                        for entry_index, entry in enumerate(old_entries):
                            for group in entry:
                                temp[entry_index] += group
                        self.problems.append(("Old Format (.cex) Style Entry Used",
                                              "line#: {}".format(index-1),
                                              interval,
                                              temp))

                    # e.g. -  someword &=d_y_MOT&=w4_50
                    if joined_entry_wrdcount:
                        temp = [""] * len(joined_entry_wrdcount)

                        for entry_index, entry in enumerate(joined_entry_wrdcount):
                            for group in entry:
                                temp[entry_index] += group
                        self.problems.append(("Entry and Word Count are Joined",
                                              "line#: {}".format(index-1),
                                              interval,
                                              temp))

                    # everything except the &= is formatted correctly
                    if just_ampersand:
                        temp = [""] * len(just_ampersand)

                        for entry_index, entry in enumerate(just_ampersand):
                            for group in entry:
                                temp[entry_index] += group
                        self.problems.append(("Ampersand Issue (\"&\" should be \"&=\")",
                                              "line#: {}".format(index-1),
                                              interval,
                                              temp))

                    # there was a "-" used instead of an "_"
                    # (either both of them, or just first/second)
                    if dash_not_underscore_all:
                        temp = [""] * len(dash_not_underscore_all)

                        for entry_index, entry in enumerate(dash_not_underscore_all):
                            for group in entry:
                                temp[entry_index] += group

                        self.problems.append(("Dash Used in Place of Underscore",
                                              "line#: {}".format(index-1),
                                              interval,
                                              temp))

                    if dash_not_underscore_first:
                        temp = [""] * len(dash_not_underscore_first)

                        for entry_index, entry in enumerate(dash_not_underscore_first):
                            for group in entry:
                                temp[entry_index] += group

                        self.problems.append(("Dash Used in Place of Underscore",
                                              "line#: {}".format(index-1),
                                              interval,
                                              temp))

                    if dash_not_underscore_second:
                        temp = [""] * len(dash_not_underscore_second)

                        for entry_index, entry in enumerate(dash_not_underscore_second):
                            for group in entry:
                                temp[entry_index] += group
                        self.problems.append(("Dash Used in Place of Underscore",
                                              "line#: {}".format(index-1),
                                              interval,
                                              temp))

                    if missing_code_just_word:
                        temp = [""] * len(missing_code_just_word)

                        for entry_index, entry in enumerate(missing_code_just_word):
                            for group in entry:
                                temp[entry_index] += group
                        self.problems.append(("Missing codes: &=x_x_XXX",
                                              "line#: {}".format(index-1),
                                              interval,
                                              temp))

                    if missing_code_just_word_zero_period:
                        temp = [""] * len(missing_code_just_word_zero_period)

                        for entry_index, entry in enumerate(missing_code_just_word_zero_period):
                            for group in entry:
                                temp[entry_index] += group
                        self.problems.append(("Missing codes: &=x_x_XXX",
                                              "line#: {}".format(index-1),
                                              interval,
                                              temp))


                    if missing_code_just_word_andcount:
                        temp = [""] * len(missing_code_just_word_andcount)

                        for entry_index, entry in enumerate(missing_code_just_word_andcount):
                            for group in entry:
                                temp[entry_index] += group
                        self.problems.append(("Missing code: &=x_x_XXX",
                                              "line#: {}".format(index-1),
                                              interval,
                                              temp))

                    if one_missing_code_first:
                        temp = [""] * len(one_missing_code_first)

                        for entry_index, entry in enumerate(one_missing_code_first):
                            for group in entry:
                                temp[entry_index] += group
                        self.problems.append(("Missing the first code: &=_x_XXX",
                                              "line#: {}".format(index-1),
                                              interval,
                                              temp))

                    if one_missing_code_second:
                        temp = [""] * len(one_missing_code_second)

                        for entry_index, entry in enumerate(one_missing_code_second):
                            for group in entry:
                                temp[entry_index] += group
                        self.problems.append(("Missing the second code: &=x__XXX",
                                              "line#: {}".format(index-1),
                                              interval,
                                              temp))

                    if one_missing_code_third:
                        temp = [""] * len(one_missing_code_third)

                        for entry_index, entry in enumerate(one_missing_code_third):
                            for group in entry:
                                temp[entry_index] += group
                        self.problems.append(("Missing the third code: &=x_x_",
                                              "line#: {}".format(index-1),
                                              interval,
                                              temp))

                    if one_missing_code_third_joined_count:
                        temp = [""] * len(one_missing_code_third_joined_count)

                        for entry_index, entry in enumerate(one_missing_code_third_joined_count):
                            for group in entry:
                                temp[entry_index] += group
                        self.problems.append(("Missing the third code and joined with word count: &=x_x_&=w",
                                              "line#: {}".format(index-1),
                                              interval,
                                              temp))


                    if missing_underscore_first:
                        temp = [""] * len(missing_underscore_first)

                        for entry_index, entry in enumerate(missing_underscore_first):
                            for group in entry:
                                temp[entry_index] += group
                        self.problems.append(("Missing the first underscore: &=xx_XXX",
                                              "line#: {}".format(index-1),
                                              interval,
                                              temp))

                    if missing_underscore_second:
                        temp = [""] * len(missing_underscore_second)

                        for entry_index, entry in enumerate(missing_underscore_second):
                            for group in entry:
                                temp[entry_index] += group
                        self.problems.append(("Missing the second underscore: &=x_xXXX",
                                              "line#: {}".format(index-1),
                                              interval,
                                              temp))

                    # correctly formatted entries
                    if entries:
                        if self.skipping:
                            logging.warning(
                                f"Object word was found in a skip region. Fix this in the .cha file.\n"
                                f"Line# {str(index)}\n"
                                f"line: {line}")
                            continue
                        else:
                            for entry in entries:
                                self.words.append([line[0:4],
                                                   entry[0],            # word
                                                   entry[3],            # utterance_type
                                                   entry[5],            # object_present
                                                   entry[7],            # speaker
                                                   entry[9],            # annotid
                                                   curr_interval[0],    # onset
                                                   curr_interval[1]])   # offset

                    last_line = line

                # intervals spanning more than 1 line start with a tab (\t)
                if line.startswith("\t"):
                    interval_reg_result = self.interval_regx.search(line)

                    if interval_reg_result is None:
                        multi_line += line
                        continue

                    prev_interval[0] = curr_interval[0]
                    prev_interval[1] = curr_interval[1]

                    # set the new curr_interval
                    interval_str = interval_reg_result.group().replace("\025", "")
                    interval = interval_str.split("_")
                    curr_interval[0] = int(interval[0])
                    curr_interval[1] = int(interval[1])

                    entries = self.entry_regx.findall(multi_line + line)

                    # check for all the possible malformed entry types
                    old_entries = self.old_entry_regx.findall(line)
                    joined_num = self.joined_num_regx.findall(line)
                    joined_entry_wrdcount = self.joined_entry_wrdcount.findall(line)
                    just_ampersand = self.just_ampersand_regx.findall(line)
                    dash_not_underscore_all = self.dash_not_underscore_all.findall(line)
                    dash_not_underscore_first = self.dash_not_underscore_first.findall(line)
                    dash_not_underscore_second = self.dash_not_underscore_second.findall(line)

                    # e.g. - someword &=d_y_MOT0 .
                    if joined_num:
                        temp = [""] * len(joined_num)

                        for entry_index, entry in enumerate(joined_num):
                            for group in entry:
                                temp[entry_index] += group
                        self.problems.append(("Joined Number and Speaker Code",
                                              "line#: {}".format(index-1),
                                              interval,
                                              temp))


                    # e.g. - someword &d|y|MOT
                    if old_entries:
                        temp = [""] * len(old_entries)

                        for entry_index, entry in enumerate(old_entries):
                            for group in entry:
                                temp[entry_index] += group
                        self.problems.append(("Old Format (.cex) Style Entry Used",
                                              "line#: {}".format(index-1),
                                              interval,
                                              temp))


                    # e.g. -  someword &=d_y_MOT&=w4_50
                    if joined_entry_wrdcount:
                        temp = [""] * len(joined_entry_wrdcount)

                        for entry_index, entry in enumerate(joined_entry_wrdcount):
                            for group in entry:
                                temp[entry_index] += group
                        self.problems.append(("Entry and Word Count are Joined",
                                              "line#: {}".format(index-1),
                                              interval,
                                              temp))


                    # everything except the &= is formatted correctly
                    if just_ampersand:
                        temp = [""] * len(just_ampersand)

                        for entry_index, entry in enumerate(just_ampersand):
                            for group in entry:
                                temp[entry_index] += group
                        self.problems.append(("Ampersand Issue (\"&\" should be \"&=\")",
                                              "line#: {}".format(index-1),
                                              interval,
                                              temp))


                    # there was a "-" used instead of an "_"
                    # (either both of them, or just first/second)
                    if dash_not_underscore_all:
                        temp = [""] * len(dash_not_underscore_all)

                        for entry_index, entry in enumerate(dash_not_underscore_all):
                            for group in entry:
                                temp[entry_index] += group
                        self.problems.append(("Dash Used in Place of Underscore",
                                              "line#: {}".format(index-1),
                                              interval,
                                              temp))

                    if dash_not_underscore_first:
                        temp = [""] * len(dash_not_underscore_first)

                        for entry_index, entry in dash_not_underscore_first:
                            for group in entry:
                                temp[entry_index] += group
                        self.problems.append(("Dash Used in Place of Underscore",
                                              "line#: {}".format(index-1),
                                              interval,
                                              temp))

                    if dash_not_underscore_second:
                        temp = [""] * len(dash_not_underscore_second)

                        for entry_index, entry in enumerate(dash_not_underscore_second):
                            for group in entry:
                                temp[entry_index] += group
                        self.problems.append(("Dash Used in Place of Underscore",
                                              "line#: {}".format(index-1),
                                              interval,
                                              temp))

                    if entries:
                        if self.skipping:
                            logging.warning(
                                f"Object word was found in a skip region. Fix this in the .cha file.\n"
                                f"Line# {str(index)}\n"
                                f"line: {line}")
                            continue
                        for entry in entries:
                            self.words.append([last_line[0:4],
                                               entry[0],            # word
                                               entry[3],            # utterance_type
                                               entry[5],            # object_present
                                               entry[7],            # speaker
                                               entry[9],            # annotid
                                               curr_interval[0],    # onset
                                               curr_interval[1]])   # offset

                    multi_line = "" # empty the mutiple line buffer

        self.phonetic_transcriptions = extract_chi_phonetic_transcriptions(self.input_file)

        if self.problems:
            # showwarning("Mistakes Found",
            #             "Fix the mistakes listed in the {} file"
            #             .format(os.path.split(self.error_file)[1]))
            self.output_problems()

    def _get_transcription(self, word):
        """
        Get phonetic transcription for a word entry in self.words
        :param word: a list from self.words
        :return: transcription if there is one, empty string otherwise
        """
        annotids, transcriptions = self.phonetic_transcriptions
        annotid = word[5]
        try:
            annotid_index = annotids.index(annotid)
            return transcriptions[annotid_index]
        except ValueError:
            return ""

    def export(self):

        comment_queue = collections.deque(self.plain_comments)
        if comment_queue:
            curr_comment = comment_queue.popleft()
        else:
            # if there are no comments, just set a dummy variable
            curr_comment = ("no comment", 0, 0)
        with open(self.output_file, "w") as output:
            writer = csv.writer(output)
            writer.writerow(["tier","word","utterance_type","object_present","speaker","annotid","timestamp","basic_level","comment","pho"])
            for entry in self.words:
                transcription = self._get_transcription(word=entry)

                # check to make sure there are comments left on the queue
                # If the current interval has passed the current comment interval,
                # pop the next comment off the queue.
                com = entry[6]
                if comment_queue:
                    if com > curr_comment[1]:
                        curr_comment = comment_queue.popleft()

                if com == curr_comment[1]:
                    writer.writerow([entry[0],
                                    entry[1],
                                    entry[2],
                                    entry[3],
                                    entry[4],
                                    entry[5],
                                    "{}_{}".format(entry[6], entry[7]),
                                    " ",
                                    curr_comment[0],
                                     transcription])

                else:
                    writer.writerow([entry[0],
                                    entry[1],
                                    entry[2],
                                    entry[3],
                                    entry[4],
                                    entry[5],
                                    "{}_{}".format(entry[6], entry[7]),
                                    " ",
                                    "NA",
                                     transcription])

        logging.info("\n\nTotal # of words: {}\n".format(len(self.words)))


    def filter_comments(self):
        """
        Filter out all the subregion and silence comments,
        leaving behind only the actual comments
        """
        for comment in self.comments:
            if ("subregion" not in comment[0]) and\
                    ("silence" not in comment[0]):
                self.plain_comments.append(comment)

    def check_intervals(self):
        """
        Makes sure there is at most 1 timestamp interval on a line
        :return: returns false if we found a problem
        """

        with open(self.input_file, "rU") as input:
            for index, line in enumerate(input):
                regx_result = self.interval_regx.findall(line)
                if regx_result:
                    if len(regx_result) > 1:
                        logging.warning(
                            f"Found more than 1 interval on a single CLAN line.\n"
                            f"Line# {str(index)}\n"
                            f"line: {line}")
                        return False
                    else:
                        continue
            return True

    def output_problems(self):
        with open(self.error_file, "w") as error_file:
            for error in self.problems:
                for element in error:
                    error_file.write(str(element) + "  ")
                error_file.write("\n\n")

    def check_personal_info_comment(self, comment, index, interval):
        if "begin" in comment or "start" in comment:
            if "begin personal information:" not in comment:
                self.problems.append(("Malformed personal info comment: line# {}"
                                      .format(index),
                                      interval,
                                      [comment]))
        if "end" in comment:
            if "end personal information" not in comment:
                self.problems.append(("Malformed personal info comment: line# {}"
                                     .format(index),
                                     interval,
                                     [comment]))

        self.curr_personal_block = PersonalInfoGroup(index)

    def parse_scrub_tier(self, line, index, interval):
        if "Scrub" not in line:
            self.problems.append(("Personal info (scrubbing) tier missing word \"Scrub\": line# {}"
                                  .format(index),
                                  interval,
                                  [line]))

        self.curr_personal_block.start_time = interval[0]
        self.curr_personal_block.end_time = interval[1]
        self.curr_personal_block.tier_line = index


class PersonalInfoGroup:
    def __init__(self, begin_com_index):
        self.start_time = None
        self.end_time = None

        self.tier_line = None
        self.start_line = begin_com_index
        self.end_line = None

        self.start_comment = ""
        self.end_comment = ""


def _extract_chi_annotation_ids_from_cha_line(cha_line: str):
    """
    Extract annotations ids from a line from a CHA file. Only the annotations uttered by CHI are extracted. We'll assume
    that an annotation id is never split over multiple lines.
    :param cha_line: a line from a cha file
    :return: list of annotation ids
    """
    return re.findall(r'_CHI_(?P<annotid>0x[a-z0-9]{6})', cha_line)


# TODO: this can be done while parsing CHA but incorporating this into the Parser class above is rather complicated and
#  so should be done after cleaning the Parser code first
def extract_chi_phonetic_transcriptions(cha_path: Path):
    """
    Extracts all phonetic transcriptions from a CHA file and the corresponding annotation ids. If a word is missing a
    transcription, it will not be included in the output. If the number of transcriptions is not equal to the number of
    CHI-uttered words in the same main tier, a ValueError will be raised. If there is a phonetic transcription, but no
    CHI-uttered words, we let it go.
    :param cha_path: path to the Cha file
    :return: annotids, transcriptions - lists of annotation ids and transcriptions
    """
    with cha_path.open('r') as f:
        lines = f.readlines()

    annotids, transcriptions = list(), list()
    # The main tier can be spread over multiple lines, the first one starting with "*XXX:\t" and the other ones - just
    # with "\t". main_tier_lines will contain all the lines in the last main tier - one or more. in_main_tier flag will
    # let us tell whether we can expect lines starting with a '\t'
    main_tier_lines = list()
    in_main_tier = False
    for line in lines:
        # First line of a main tier
        if line.startswith('*'):
            main_tier_lines = [line]
            in_main_tier = True

        # A continuation line of a main tier
        elif in_main_tier and line.startswith('\t'):
            main_tier_lines.append(line)

        # Not a main tier line
        else:
            in_main_tier = False
            if line.startswith('%pho:\t'):
                annotids_ = [annotid
                             for cha_line in main_tier_lines
                             for annotid in _extract_chi_annotation_ids_from_cha_line(cha_line)]
                # Remove tier type and then tab on the left and a line end on the right
                content = line.replace('%pho:', '').rstrip()
                transcriptions_ = content.split()

                # The numbers of words and transcriptions should be equal. But if there are no annotated words but there
                # are still some transcription, we will ignore it.
                if len(annotids_) != len(transcriptions_) and not (len(annotids_) == 0):
                    raise ValueError('The numbers of annotation ids and the phonetic transcriptions do not match')

                annotids.extend(annotids_)
                transcriptions.extend(transcriptions_)

    return annotids, transcriptions


def export_cha_to_csv(cha_path, output_folder):
    """
    Runs parse_clan2 (a version of it) on a cha file at cha_path and outputs the results to a csv_path
    :param cha_path: Path
    :param output_folder: Path to folder that will contain the _processed.csv and _errors.csv files. If None, output is
    saved to the save folder where cha_path is.
    :return: Path|None, if there were known errors, return the error file path. If the file could not be parsed, return
    cha_path
    """
    try:
        # Parser parses implicitly
        parser = Parser(input_path=cha_path, output=output_folder)
        error_file_path = Path(parser.error_file)
        if error_file_path.exists():
            return error_file_path
    except Exception:
        return cha_path


"""
This module checks that the LENA folder contains all and only files that we expect to be there.
"""
import re
from enum import auto, unique
from pathlib import Path

import pandas as pd
from strenum import StrEnum

from ..paths import get_lena_path, compose_full_recording_id, parse_full_recording_id, get_lena_recording_path, get_vihi_path

# Strings to match literally, and re.Pattern objects to match in the sense of re.match. Add $ to match the full name.
IGNORED = [re.compile(r'~.*\.docx$'), '.DS_Store', '.git']


@unique
class AuditStatus(StrEnum):
    ignored = auto()
    expected = auto()
    unexpected = auto()
    missing = auto()


def _name_matches(path, rule_list):
    """
    Checks if the name of the file at path matches any of the rules
    :param path:
    :param rule_list: a list of strings and re.Pattern objects to match against
    :return: rule_list element that the path name matched if there was a match, None if there were no matches
    """
    name = path.name
    for rule in rule_list:
        if isinstance(rule, str) and name == rule:
            return rule
        elif isinstance(rule, re.Pattern) and rule.match(name):
            return rule
    return


def audit_recording_folder(folder_path: Path, population: str, subject_id: str, recording_id: str):
    """
    Checks a recording folder for
    :param folder_path: path to the folder with the recording
    :param population: population
    :param subject_id: subject id string
    :param recording_id: recording id string
    :return: a dataframe with the status of each file in folder in `path`
    """
    assert folder_path.is_dir()
    full_recording_id = compose_full_recording_id(population, subject_id, recording_id)
    assert folder_path.name == full_recording_id

    expected_objects = [
        f'{full_recording_id}.eaf',
        f'{full_recording_id}.its',
        f'{full_recording_id}.wav',
        f'{full_recording_id}.pfsx',
        f'{full_recording_id}.upl',
        f'{full_recording_id}_lena5min.csv',
        f'VIHI_Coding_Issues_{full_recording_id}.docx'
    ]

    objects_statuses = list()
    satisfied_rules = list()
    folder_contents = [path.relative_to(folder_path) for path in folder_path.iterdir()]
    for path in folder_contents:
        if _name_matches(path, rule_list=IGNORED):
            objects_statuses.append(AuditStatus.ignored)
            continue

        rule_matched = _name_matches(path, rule_list=expected_objects)
        if rule_matched:
            satisfied_rules.append(rule_matched)
            objects_statuses.append(AuditStatus.expected)
        else:
            objects_statuses.append(AuditStatus.unexpected)

    missing_files = [rule for rule in expected_objects if rule not in satisfied_rules]
    audit_results = ([(path.as_posix(), str(status)) for path, status in zip(folder_contents, objects_statuses)]
                     + [(rule, str(AuditStatus.missing)) for rule in missing_files])

    return (pd.DataFrame(columns=['relative_path', 'status'], data=audit_results)
            .sort_values(by=['status', 'relative_path'])
            .reset_index(drop=True))


def audit_all_recordings():
    """
    Checks:
    - all the folders at levels between ".../LENA" and the recording-level folders,
    - all the recording-level folders using `audit_recording_folder`
    :return:
    """
    recordings_list = pd.read_csv(get_vihi_path().joinpath('vihi_data_check', 'recordings.csv'))
    recordings_list[['population', 'subject_id', 'recording_id']] = pd.DataFrame(
        recordings_list.recording.apply(parse_full_recording_id).to_list())

    # Check the folder presence at population, subject, and recording levels
    lena_dir = get_lena_path()
    expected_folders = {
        folder
        for population, subject_id, recording_id
        in recordings_list[['population', 'subject_id', 'recording_id']].drop_duplicates().to_records(index=False)
        for folder in (
            lena_dir / population,
            lena_dir / population / f'{population}_{subject_id}',
            lena_dir / population / f'{population}_{subject_id}' / f'{population}_{subject_id}_{recording_id}'
        )}

    actual_folders = {
        path
        for glob_results in (lena_dir.glob('*'),
                             lena_dir.glob('*/*'),
                             lena_dir.glob('*/*/*'))
        for path in glob_results
        if path.is_dir()
    }

    folder_statuses = list()
    unexpected = actual_folders - expected_folders
    expected = actual_folders & expected_folders
    missing = expected_folders - actual_folders
    for folder_list, status in [(unexpected, AuditStatus.unexpected),
                                (expected, AuditStatus.expected),
                                (missing, AuditStatus.missing)]:
        for folder in folder_list:
            folder_statuses.append(dict(parent=lena_dir,
                                        relative_path=folder.relative_to(lena_dir), status=status))

    # Check the recording folders contents
    recordings_list['folder_path'] = recordings_list.apply(
        lambda row: get_lena_recording_path(row.population, row.subject_id, row.recording_id),
        axis='columns')
    recording_audits = recordings_list.apply(
        lambda row:
        audit_recording_folder(
            folder_path=row.folder_path,
            population=row.population,
            subject_id=row.subject_id,
            recording_id=row.recording_id).assign(parent=row.folder_path),
        axis='columns',
        result_type='reduce')

    # Combine the results
    full_results = pd.concat([pd.DataFrame(folder_statuses)] + recording_audits.to_list())
    assert full_results.columns.to_list() == ['parent', 'relative_path', 'status']

    # Sort by full path and convert "parent" to string
    full_results = (full_results
                    .assign(full_path=lambda df: df.parent / df.relative_path)
                    .sort_values(by='full_path')
                    .drop(columns='full_path')
                    .assign(parent=lambda df: df.parent.apply(lambda path: path.as_posix()))
                    )

    return full_results



import os
from pathlib import Path

# Environment variables
PN_OPUS = 'PN_OPUS_PATH'


def get_pn_opus_path():
    """
    Tries to find PN-OPUS at the path set via an environment variable.
    If it is not set, looks at the standard mount point on MacOS: /Volumes/pn-opus
    :return: Path object pointing to the PN-OPUS network drive if found
    :raises: ValueError if drive not found
    """
    default_path = Path('/Volumes/pn-opus/')
    if PN_OPUS in os.environ:
        path = Path(os.environ[PN_OPUS])
    else:
        path = default_path

    if path.exists():
        return path
    else:
        raise ValueError(f'Could not locate PN-OPUS at {path}, check that VPN is connected and PN-OPUS is mounted.\n'
                         f'If this is not the correct location, set/update environment/shell variable {PN_OPUS}')



import pytest

from blabpy.seedlings.paths import get_seedlings_path, _get_home_visit_folder, _get_coding_folder, _get_analysis_folder, \
    _check_modality, _get_annotation_path, get_opf_path, get_cha_path, _get_all_paths, get_all_opf_paths, \
    get_all_cha_paths, MISSING_AUDIO_RECORDINGS, MISSING_VIDEO_RECORDINGS, get_basic_level_path, \
    get_all_basic_level_paths

TEST_CHILD, TEST_MONTH = 6, 10
AUDIO = 'Audio'
VIDEO = 'Video'
ANNOTATION_FILE_COUNT = 527


def test_get_seedlings_path():
    get_seedlings_path()


@pytest.mark.parametrize(argnames='get_function',
                         argvalues=[_get_home_visit_folder, _get_coding_folder, _get_analysis_folder])
def test_get_child_month_folder(get_function):
    folder = get_function(child=TEST_CHILD, month=TEST_MONTH)
    assert folder.exists()


def test__check_modality():
    _check_modality(AUDIO)
    _check_modality(VIDEO)
    with pytest.raises(AssertionError):
        _check_modality('tactile')


@pytest.mark.parametrize(argnames='modality',
                         argvalues=[AUDIO, VIDEO])
def test__get_annotation_path(modality):
    annotation_path = _get_annotation_path(child=TEST_CHILD, month=TEST_MONTH, modality=modality)
    assert annotation_path.exists()


def test_get_opf_path():
    opf_path = get_opf_path(child=TEST_CHILD, month=TEST_MONTH)
    assert opf_path.exists()


def test_get_cha_path():
    cha_path = get_cha_path(child=TEST_CHILD, month=TEST_MONTH)
    assert cha_path.exists()


@pytest.mark.parametrize(argnames=('get_single_file_function', 'missing_child_month_combinations'),
                         argvalues=((get_cha_path, MISSING_AUDIO_RECORDINGS),
                                    (get_opf_path, MISSING_VIDEO_RECORDINGS)))
def test__get_all_paths(get_single_file_function, missing_child_month_combinations):
    paths = _get_all_paths(get_single_file_function=get_single_file_function,
                           missing_child_month_combinations=missing_child_month_combinations)
    assert len(paths) == ANNOTATION_FILE_COUNT


def test_get_all_opf_paths():
    paths = get_all_opf_paths()
    assert len(paths) == ANNOTATION_FILE_COUNT


def test_get_all_cha_paths():
    paths = get_all_cha_paths()
    assert len(paths) == ANNOTATION_FILE_COUNT


@pytest.mark.parametrize(argnames='modality',
                         argvalues=(AUDIO, VIDEO))
def test_get_basic_level_path(modality):
    path = get_basic_level_path(child=TEST_CHILD, month=TEST_MONTH, modality=modality)
    assert path.exists()


@pytest.mark.parametrize(argnames='modality',
                         argvalues=(AUDIO, VIDEO))
def test_get_all_basic_level_paths(modality):
    paths = get_all_basic_level_paths(modality=modality)
    assert len(paths) == ANNOTATION_FILE_COUNT



import pandas as pd
from pympi import Eaf


class EafPlus(Eaf):
    """
    This class is just pympi.Eaf plus a few extra methods.
    """

    def get_time_intervals(self, tier_id):
        """
        Get time slot intervals from all tiers with a given id.
        :param tier_id: string with a tier id ('code', 'context', etc.)
        :return: [(start_ms, end_ms), ...]
        """
        # From `help(pympi.Eaf)` for the `tiers` attribute:
        #
        # tiers (dict)
        #
        # Tiers, where every tier is of the form:
        # {tier_name -> (aligned_annotations, reference_annotations, attributes, ordinal)},
        # aligned_annotations of the form: [{id -> (begin_ts, end_ts, value, svg_ref)}],
        # reference annotations of the form: [{id -> (reference, value, previous, svg_ref)}].

        # We only need aligned annotations. And from those we only need begin_ts and end_ts - ids of the time slots
        # which we will convert to timestamps in ms using eaf.timeslots. .eaf files no nothing about sub-recordings,
        # so all the timestamp are in reference to the wav file.
        aligned_annotations = self.tiers[tier_id][0]
        timeslot_ids = [(begin_ts, end_ts) for begin_ts, end_ts, _, _ in aligned_annotations.values()]
        timeslots = [(self.timeslots[begin_ts], self.timeslots[end_ts]) for begin_ts, end_ts in timeslot_ids]

        return timeslots

    def get_values(self, tier_id):
        """
        Get values from a tier.
        :param tier_id:
        :return: list of values
        """
        # Same logic as in get_time_intervals
        aligned_annotations = self.tiers[tier_id][0]
        values = [value for _, _, value, _ in aligned_annotations.values()]

        return values

    def get_participant_tier_ids(self):
        participant_tier_ids = [tier_id
                                for tier_id, (_, _, attributes, _)
                                in self.tiers.items()
                                if 'PARTICIPANT' in attributes
                                and attributes['LINGUISTIC_TYPE_REF'] == 'transcription']
        return participant_tier_ids

    def _get_aligned_annotations(self, tier_id):
        time_intervals = self.get_time_intervals(tier_id=tier_id)
        if len(time_intervals) == 0:
            return None

        onsets, offsets = zip(*time_intervals)
        aligned_annotations, _, _, _ = self.tiers[tier_id]
        if aligned_annotations:
            ids, annotations = zip(*[(id_, annotation)
                                     for id_, (_, _, annotation, _)
                                     in aligned_annotations.items()])
            return pd.DataFrame.from_dict(dict(
                onset=onsets, offset=offsets,
                annotation=annotations,
                annotation_id=ids))
        else:
            return pd.DataFrame(columns=['onset', 'offset', 'annotation', 'annotation_id'])

    def _get_reference_annotations(self, tier_id):
        _, reference_annotations, _, _ = self.tiers[tier_id]
        if reference_annotations:
            parent_ids, daughter_ids, annotations = zip(*[
                (parent_id, daughter_id, annotation)
                for daughter_id, (parent_id, annotation, _, _)
                in reference_annotations.items()])
            return pd.DataFrame.from_dict({
                'annotation': annotations,
                'annotation_id': daughter_ids,
                'parent_annotation_id': parent_ids,})
        else:
            return pd.DataFrame(columns=['annotation', 'annotation_id', 'parent_annotation_id'])

    def get_full_annotations_for_participant(self, tier_id):
        """
        Return annotations for a given participant tier, including daughter annotations (vcm, lex, ...)
        :param tier_id: participant's tier id
        :return: pd.DataFrame with columns onset, offset, annotation, xds (non-chi), or vcm, lex, and mwu (CHI)
        """
        annotations_df = self._get_aligned_annotations(tier_id=tier_id)
        if annotations_df is None:
            return None

        # as we add daughter tiers, deepest_annotation_id column will change to their ids
        annotations_df = annotations_df.rename(columns={'annotation_id': 'deepest_annotation_id'})
        n_annotations = annotations_df.shape[0]

        if tier_id == 'CHI':
            daughter_tier_types = ('vcm', 'lex', 'mwu')
        else:
            daughter_tier_types = ('xds', )

        for daughter_tier_type in daughter_tier_types:
            daughter_tier_id = f'{daughter_tier_type}@{tier_id}'
            daughter_annotations_df = self._get_reference_annotations(tier_id=daughter_tier_id)
            daughter_annotations_df = daughter_annotations_df.rename(columns={'annotation': daughter_tier_type})
            # Merge with previously extracted annotations
            annotations_df = (annotations_df.merge(daughter_annotations_df,
                                 how='left',
                                 left_on='deepest_annotation_id',
                                 right_on='parent_annotation_id')
             .drop(columns=['deepest_annotation_id', 'parent_annotation_id'])
             .rename(columns={'annotation_id': 'deepest_annotation_id'})
             )

        assert annotations_df.shape[0] == n_annotations
        return annotations_df.drop(columns=['deepest_annotation_id'])

    def get_full_annotations(self):
        """
        All participant-tier annotations, including daughter tiers (xds, vcm, ...)
        :return: pd.DataFrame with columns participant, onset, offset, annotation, xds ,vcm, lex, and mwu
        """
        participant_tier_ids = self.get_participant_tier_ids()
        all_annotations = [self.get_full_annotations_for_participant(tier_id=participant_tier_id)
                           for participant_tier_id in participant_tier_ids]
        all_annotations_df = (
            pd.concat(objs=all_annotations,
                      keys=participant_tier_ids,
                      names=['participant', 'order'])
            .reset_index('participant', drop=False)
            .reset_index(drop=True))

        return all_annotations_df.sort_values(by=['onset', 'offset', 'participant'])



"""
This module contains function that interact with files - either reading or writing them. The idea is that all the
other function are supposed to be pure. That is not true at all at this time though.
"""
from pathlib import Path

import numpy as np
import pandas as pd

from .intervals.intervals import add_metric, make_intervals, add_annotation_intervals_to_eaf, _region_output_files, \
    select_best_intervals, _extract_interval_info, INTERVALS_FOR_ANNOTATION_COUNT, INTERVALS_EXTRA_COUNT
from ..its import Its, ItsNoTimeZoneInfo
from .paths import get_its_path, parse_full_recording_id, get_eaf_path, get_rttm_path
from ..utils import df_to_list_of_tuples
from ..vtc import read_rttm
from ..eaf import EafPlus


def gather_recordings(full_recording_id, forced_timezone=False):
    """
    Gets sub-recordings for a given single recording. See Its.gather_recordings for details.
    """
    its_path = get_its_path(**parse_full_recording_id(full_recording_id))
    its = Its.from_path(its_path, forced_timezone=forced_timezone)
    try:
        return its.gather_recordings()
    except ItsNoTimeZoneInfo as e:
        raise ItsNoTimeZoneInfo(f'No timezone info in \n{its_path}') from e


# TODO: these function don't need to exist. Just add full_recording_id as an alternative argument to all the get_*
#  functions in vihi.paths. It should be part of the future decorator that checks the arguments (see issue #18).
def get_vtc_data(full_recording_id):
    """
    Get VTC data for a given recording.
    :param full_recording_id: a string of '{population}_{subject}_{recording_id}' format.
    :return: a pandas dataframe with all the data from the VTC outputs (.rttm files).
    """
    rttm_path = get_rttm_path(**parse_full_recording_id(full_recording_id))
    return read_rttm(rttm_path)


def get_eaf_path_from_full_recording_id(full_recording_id):
    """
    Find annotations eaf for a given recording.
    :param full_recording_id:
    :return:
    """
    return get_eaf_path(**parse_full_recording_id(full_recording_id))


class NoPreviousEaf(Exception):
    pass


def _load_eaf(full_recording_id: Path):
    """
    Loads eaf if it exists, otherwise throws an error
    :param full_recording_id: full recording id
    :return: a blabpy.eaf.EafPlus object
    :raises: NoPreviousEaf
    """
    eaf_path = get_eaf_path_from_full_recording_id(full_recording_id)
    try:
        eaf = EafPlus(eaf_path)
    except FileNotFoundError as e:
        raise NoPreviousEaf(f'There is no eaf file where we expected to find it:\n{eaf_path}') from e

    return eaf


# TODO: this function should handle both random and high-volubility sampling
def add_intervals_for_annotation(full_recording_id, forced_timezone=None):
    """
    For a given recording, finds the intervals that maximize vtc_total_speech_duration - total duration of all speech
    segments as classified by VTC.
    Then adds them to the corresponding eaf file.
    :param forced_timezone: in case where timezone info was removed from its or its has incorrect info. See Its.__init__
    :param full_recording_id: full recording id, string
    :return: None
    """
    # Prepare intervals to select from
    intervals = make_intervals(sub_recordings=gather_recordings(full_recording_id, forced_timezone=forced_timezone))
    vtc_data = get_vtc_data(full_recording_id)
    intervals = add_metric(intervals=intervals, vtc_data=vtc_data)

    # Load existing intervals
    eaf = _load_eaf(full_recording_id)
    existing_intervals = _extract_interval_info(eaf)
    if len(existing_intervals) > 15:
        raise ValueError(f'{full_recording_id} has more than 15 intervals already. Likely, you are adding intervals'
                         f' to it for the second time.')
    existing_code_intervals = df_to_list_of_tuples(existing_intervals[['code_onset_wav', 'code_offset_wav']])

    # Select intervals that maximize vtc_total_speech_duration
    best_intervals = select_best_intervals(
        intervals,
        existing_code_intervals=existing_code_intervals,
        n_to_select=(INTERVALS_FOR_ANNOTATION_COUNT + INTERVALS_EXTRA_COUNT))
    sampling_types = np.where(
        # Intervals with the highest value of the metric will be used in the main part, the next
        # INTERVALS_EXTRA_COUNT ones - in the extra part
        best_intervals.metric_value.rank(method='first', ascending=False)
        <= INTERVALS_FOR_ANNOTATION_COUNT,
        'high-volubility',
        'high-volubility-extra')
    best_intervals.insert(0, 'sampling_type', sampling_types)
    eaf, best_intervals = add_annotation_intervals_to_eaf(eaf, best_intervals)

    # Save eaf
    output_file_paths = _region_output_files(full_recording_id=full_recording_id)
    eaf.to_file(output_file_paths['eaf'])

    # Save log
    old_log = pd.read_csv(output_file_paths['csv'])
    best_intervals.insert(0, 'full_recording_id', full_recording_id)
    # TODO: add selected_regions to the test
    selected_regions = pd.concat([old_log, best_intervals], ignore_index=True)
    selected_regions.to_csv(output_file_paths['csv'], index=False)



"""
Functions to work with the VTC output files.
These files have an ".rttm" extension, they are space-separated text fiels without column names.

VTC - Voice Type Classifier - a set of of voice classification models and the code to apply them.
See https://github.com/MarvinLvn/voice-type-classifier for details.
See GitBook for instructions on how to run VTC on the Duke cluster.
"""

from pathlib import Path

import pandas as pd


RTTM_COLUMNS = ['SPEAKER', 'filename', 'column3',
                'onset', 'duration', 'column6', 'column7',
                'voice_type', 'column9', 'column10']


def read_rttm(path: Path):
    """
    Reads an rttm file output by VTC
    :param path: path to the rttm file
    :return: a pandas dataframe object
    """
    return pd.read_csv(path, delimiter=' ', names=RTTM_COLUMNS, dtype=str, na_filter=False)


def _write_rttm(df: pd.DataFrame, output_path: Path):
    """
    Writes rttm dataframe to a file
    :param df: a pandas dataframe
    :param output_path: output path
    :return: None
    """
    assert df.columns.tolist() == RTTM_COLUMNS
    output_path.parent.mkdir(exist_ok=True, parents=True)
    df.to_csv(output_path, sep=' ', index=False, columns=RTTM_COLUMNS, header=False)


def split_rttm(rttm_path: Path, output_dir: Path):
    """
    Splits VTC's rttm output file into separate files corresponding to the individual input files.

    When VTC is run on a folder with wav files, it will create a single output file with the filename column which is
    not always convenient.

    If some of the files already exist and their contents are different, an error will be raised and nothing will be
    done with any of the files.
    :param rttm_path: path to the common rttm file
    :param output_dir: separate files will be output into output_dir/<filename>/<input file name>
    :return: None

    Sample code:
    from pathlib import Path
    from blabpy.vtc import split_rttm
    split_rttm(Path('all.rttm'), Path('separated'))
    """
    rttm_grouped = read_rttm(rttm_path).groupby('filename')

    def _output_path(filename_):
        return output_dir / filename_ / rttm_path.name

    # Check that the output files either do not exist or contain the same information
    for filename, sub_df in rttm_grouped:
        output_path = _output_path(filename)
        if output_path.exists():
            assert read_rttm(output_path).equals(sub_df.reset_index(drop=True)), (
                f'The following file already exists and contains different information from the input file:'
                f'{output_path}')

    # Write the individual files
    for filename, sub_df in rttm_grouped:
        output_path = _output_path(filename)
        if not output_path.exists():
            _write_rttm(sub_df, _output_path(filename))



from csv import QUOTE_NONNUMERIC

import pandas as pd
import numpy as np

from . import VALID_UTTERANCE_TYPE_CODES, VALID_OBJECT_PRESENT_CODES
from .paths import _check_modality, get_all_basic_level_paths, VIDEO, AUDIO


# Maps columns names in basic level files to standardized names.
COLUMN_NAME_MAPPER = {
    # For video, simply remove 'labeled_object.' and change 'id' to 'annotid'
    VIDEO: lambda column_name: (column_name.replace('labeled_object.', '')
                                if column_name != 'labeled_object.id' else 'annotid'),
    # For audio, rename 'word' to 'object'
    AUDIO: lambda column_name: 'object' if column_name == 'word' else column_name
}

COLUMNS_BY_MODALITY = {
    AUDIO: ["tier", "object", "utterance_type", "object_present",
            "speaker", "timestamp", "basic_level", "annotid", 'pho'],
    VIDEO: ["ordinal", "onset", "offset",
            "object", "utterance_type", "object_present",
            "speaker", "basic_level", "annotid", 'pho'],
    'combined': ['ordinal',
                 'onset',
                 'offset',
                 'object',
                 'utterance_type',
                 'object_present',
                 'speaker',
                 'basic_level',
                 'annotid',
                 'id',
                 'subj',
                 'month',
                 'SubjectNumber',
                 'audio_video',
                 'tier',
                 'pho']
}


def load_and_normalize_column_names(basic_level_path, modality):
    """
    Loads individual basic level file and normalizes its column names. For audio, additionally splits timestamp column
    into onset and offset columns.
    :param basic_level_path: path to a basic level file
    :param modality: Audio/Video
    :return:
    """
    _check_modality(modality)
    df = (pd
          .read_csv(basic_level_path)
          .rename(columns=COLUMN_NAME_MAPPER[modality])
          [COLUMNS_BY_MODALITY[modality]]
          .assign(id=basic_level_path.name))

    # Each modality requires a bit of additional manipulation
    if modality == AUDIO:
        # split timestamp (single string separated by "_", e.g. 4567_4589) into onset and offset (4567, 4589)
        df[['onset', 'offset']] = df.timestamp.str.split('_', expand=True).astype(int)
        df.drop(columns='timestamp', inplace=True)

    if modality == VIDEO:
        # set "ordinal" to a type that support missing integer values (by default, pandas can't handle int and NAs in
        # one coumn
        df['ordinal'] = df.ordinal.astype(pd.Int64Dtype())

    # Some basic level data have trailing whitespace which additionally results in "NA " not being recognized as nan
    df['basic_level'] = df.basic_level.str.strip().replace('NA', np.nan)

    return df


def gather_basic_level_annotations(modality):
    """
    Combines all individual child-month basic level files (i.e.,
     <Modality>_Analysis/<child>_<month>_<modality>_sparse_code.csv) into a single dataframe.
    That is, it output half of all_basiclevel.csv.
    :param modality: Audio/Video
    :return: a pandas DataFrame
    """
    _check_modality(modality)
    basic_level_paths = get_all_basic_level_paths(modality=modality)
    return pd.concat([load_and_normalize_column_names(basic_level_path, modality=modality)
                      for basic_level_path in basic_level_paths])


def _combine_basic_level_annotations(all_audio_df, all_video_df):
    # Concatenate
    all_df = pd.concat(objs=[all_video_df, all_audio_df],
                       keys=['video', 'audio'],
                       names=['audio_video', 'index']
                       ).reset_index(0)

    # Add extra columns
    all_df[['subj', 'month']] = all_df.id.str.split('_', expand=True)[[0, 1]]
    all_df['SubjectNumber'] = all_df.subj + '_' + all_df.month

    # Enforce column order
    all_df = all_df[COLUMNS_BY_MODALITY['combined']]

    return all_df


def gather_all_basic_level_annotations(keep_comments=False, keep_basic_level_na=False):
    """

    :param keep_comments: whether to keep the comments
    :param keep_basic_level_na: whether to keep rows where basic level was manually set to NA by an annotator, must be
    True if keep
    :return:
    """
    # Keep behavior consistent with the previous R code
    if keep_comments and not keep_basic_level_na:
        raise ValueError('When keeping comments, keep empty basic level as well')

    all_audio_df = gather_basic_level_annotations(modality=AUDIO)
    all_video_df = gather_basic_level_annotations(modality=VIDEO)
    all_df = _combine_basic_level_annotations(all_audio_df=all_audio_df, all_video_df=all_video_df)

    # Remove comments
    if not keep_comments:
        all_df = all_df[~all_df.object.str.startswith('%com:')]

    # Remove rows without the basic level information
    if not keep_basic_level_na:
        all_df = all_df[~all_df.basic_level.isna()]

    # Sort by modality, month and subject for consistency with the older R code.
    # And then by ordinal (relevent for opfs only), onset, offset, and annotid for semi-consistency between versions
    all_df = all_df.sort_values(by=['audio_video', 'month', 'subj', 'ordinal', 'onset', 'offset', 'annotid'],
                                ascending=[False, True, True, True, True, True, True])

    # Convert a subset of the columns to factors (categorical in the pandas's terms)
    factor_columns = ['object', 'utterance_type',
                      'object_present', 'speaker', 'basic_level', 'id', 'subj',
                      'month', 'SubjectNumber', 'audio_video', 'tier']
    all_df[factor_columns] = all_df[factor_columns].astype('category')

    all_df.reset_index(drop=True, inplace=True)

    return all_df


def check_for_errors(all_basic_level_df: pd.DataFrame):
    """
    Checks the all_basic_level dataframe for:
    1. Duplicate annotation ids.
    2. Invalid "object present".
    3. Invalid utterance type.
    :param all_basic_level_df:
    :return: None if there were no errors. Otherwise, return a subset of all_basic_level_df with additional "error_type"
    column. The subset contains all the rows with errors.
    """
    df = all_basic_level_df

    # Find duplicate annotation ids
    # keep=False - mark all duplicates as such
    duplicates = df[df.duplicated(subset=['annotid'], keep=False)]

    # Invalid codes
    invalid_utterance_type = df[~df.utterance_type.isin(VALID_UTTERANCE_TYPE_CODES)]
    invalid_object_present = df[~df.object_present.isin(VALID_OBJECT_PRESENT_CODES)]

    all_errors = pd.concat(
        objs=[duplicates, invalid_utterance_type, invalid_object_present],
        keys=['duplicate annotation id', 'invalid utterance type code', 'invalid object present code'],
        names=['error_type', 'index']).reset_index(0)

    if len(all_errors.index) > 0:
        return all_errors
    else:
        return None


def write_all_basic_level_to_csv(all_basic_level_df, csv_path):
    """
    Write the output of gather_all_basic_level_annotations to a csv file in a way consistent with the older R code.
    The result is still not fully consistent but it is close enough.
    (readr::write_csv writes number 10000 as 1e+5 which was too silly to emulate)
    :param all_basic_level_df: a pandas DataFrame
    :param csv_path: a path to the csv or None if you want to return the string that would have been written
    :return:
    """
    # For consistency with readr::write_csv that quotes strings but does not quote NAs, we'll have to use the following
    # trick from https://www.reddit.com/r/Python/comments/mu65ms/quoting_of_npnan_with_csvquote_nonnumeric_in/
    # This way, NAs will be considered numeric and won't be quoted.
    na = type("NaN", (float,), dict(__str__=lambda _: "NA"))()
    return all_basic_level_df.to_csv(csv_path, index=False, quoting=QUOTE_NONNUMERIC, na_rep=na)


def write_all_basic_level_to_feather(all_basic_level_df, feather_path):
    """
    Write the output of gather_all_basic_level_annotations to a feather file.
    :param all_basic_level_df: a pandas DataFrame
    :return:
    """
    all_basic_level_df.to_feather(feather_path)



from pathlib import Path
from itertools import product

from blabpy.seedlings.pipeline import make_updated_all_basic_level_here


def test_make_updated_all_basic_level_here(tmpdir):
    """
    Only checks that all_basiclevel can be successfully created. Require connection to PN-OPUS.
    """
    with tmpdir.as_cwd():
        make_updated_all_basic_level_here()
        cwd = Path()
        for ending, extension in product(('', '_NA'), ('.csv', '.feather')):
            filename = 'all_basiclevel' + ending + extension
            assert cwd.joinpath(filename).exists()



from io import StringIO

import pandas as pd

from blabpy.vihi.intervals import templates
from blabpy.vihi.intervals.intervals import create_eaf_from_template, _extract_interval_info

from blabpy.vihi.tests.test_pipeline import TEST_FULL_RECORDING_ID

age_in_days = int(TEST_FULL_RECORDING_ID.split('_')[-1])
# TODO: check if this is correct. Here, it is fine anyway but would be good to know.
age_in_months = int(age_in_days // 30.25)
etf_template_path, pfsx_template_path = templates.choose_template(age_in_months=age_in_months)


intervals_str = """full_recording_id,sampling_type,code_onset_wav,code_offset_wav,context_onset_wav,context_offset_wav,code_num,on_off
TEST_123_290,random,353000,473000,233000,533000,1,353000_473000
TEST_123_290,random,1200000,1320000,1080000,1380000,2,1200000_1320000"""
intervals_string_i0 = StringIO(intervals_str)
intervals = pd.read_csv(intervals_string_i0)

eaf = create_eaf_from_template(etf_template_path, intervals=intervals)
eaf.to_file('expected.eaf')


import logging
import warnings
from itertools import product
from pathlib import Path

import pandas as pd

from .cha import export_cha_to_csv
from .gather import gather_all_basic_level_annotations, write_all_basic_level_to_csv, write_all_basic_level_to_feather, \
    check_for_errors
from .listened_time import listen_time_stats_for_report, RECORDINGS_WITH_FOUR_SUBREGIONS, _get_subregion_count, \
    _preprocess_region_info
from .merge import create_merged, FIXME
from .opf import export_opf_to_csv
from .paths import get_all_opf_paths, get_all_cha_paths, get_basic_level_path, _parse_out_child_and_month, \
    ensure_folder_exists_and_empty, AUDIO, VIDEO, _check_modality, get_seedlings_path

# Placeholder value for words without the basic level information
from .scatter import copy_all_basic_level_files_to_subject_files


def export_all_opfs_to_csv(output_folder: Path, suffix='_processed'):
    """
    Exports all opf files, adds suffix to their names and saves to the output_folder
    :param output_folder: Path p
    :param suffix: str
    :return:
    """
    ensure_folder_exists_and_empty(output_folder)

    opf_paths = get_all_opf_paths()
    seedlings_path = get_seedlings_path()
    for opf_path in opf_paths:
        # Add suffix before all extensions
        extensions = ''.join(opf_path.suffixes)
        output_name = opf_path.name.replace(extensions, suffix + '.csv')

        print(f'Exporting opf file at <SEEDLINGS_ROOT>/{opf_path.relative_to(seedlings_path)}')
        export_opf_to_csv(opf_path=opf_path, csv_path=(output_folder / output_name))


def export_all_chas_to_csv(output_folder: Path, log_path=Path('cha_parsing_errors_log.txt')):
    """
    Exports all cha files to output_folder
    :param output_folder: Path
    :param log_path: Path, file where errors are logged if any
    :return: Path|None, if there were any errors, returns path to the error log file
    """
    ensure_folder_exists_and_empty(output_folder)

    # These will hold paths to files with problems if any
    could_not_be_parsed = list()
    parsed_with_errors = list()

    cha_paths = get_all_cha_paths()
    seedlings_path = get_seedlings_path()
    for cha_path in cha_paths:
        print(f'Exporting cha file at <SEEDLINGS_ROOT>/{cha_path.relative_to(seedlings_path)}')
        problems = export_cha_to_csv(cha_path=cha_path, output_folder=output_folder)
        if not problems:
            continue

        # If there were any problems, take note
        if problems == cha_path:
            could_not_be_parsed.append(problems)
        else:
            parsed_with_errors.append((cha_path, problems))

    # Write errors to the log file
    if could_not_be_parsed or parsed_with_errors:
        with log_path.open('w', encoding='utf-8') as f:

            if could_not_be_parsed:
                f.write('The following files could not be parsed:\n\n')
                for path in could_not_be_parsed:
                    f.write(str(path) + '\n')

            if parsed_with_errors:
                f.write('The following files were parsed with errors:\n\n')
                for cha_path, error_path in parsed_with_errors:
                    f.write(f'Cha file: {str(cha_path.absolute())}\n')
                    f.write(f'Error log: {str(error_path.absolute())}\n')
    else:
        log_path = None

    if parsed_with_errors:
        warnings.warn(f'Some cha files were parsed with errors. For details, see:\n {str(log_path.absolute())}')

    if could_not_be_parsed:
        raise Exception(f'Some cha files could not parsed at all. Try exportint them individually. For details, see:\n'
                        f' {str(log_path.absolute())}')

    return log_path


def merge_annotations_with_basic_level(exported_annotations_folder, output_folder, mode,
                                       exported_suffix='_processed.csv'):
    """
    Merges all exported annotation files in output_folder and saves them to output_folder which must be empty.
    :param exported_annotations_folder: the input folder
    :param output_folder: the output folder
    :param mode: 'audio'|'video' - which modality these files came from
    :param exported_suffix: the ending of the exported annotation file names, needed because export_cha_to_csv exports
    two files: the actual csv and the errors file
    :return: 
    """
    ensure_folder_exists_and_empty(output_folder)

    # Find/assemble all necessary paths
    annotation_files = list(exported_annotations_folder.glob(f'*{exported_suffix}'))
    basic_level_files = [get_basic_level_path(**_parse_out_child_and_month(annotation_file), modality=mode.capitalize())
                         for annotation_file in annotation_files]
    output_files = [output_folder / basic_level_file.name for basic_level_file in basic_level_files]

    # Merge and save
    seedlings_path = get_seedlings_path()

    def _create_merged(file_new, file_old, file_merged, mode):
        """Exists to add the printing part"""
        print(f'Merging annotations in {file_new}\n'
              f'with basic level info in <SEEDLINGS_ROOT>/{file_old.relative_to(seedlings_path)}')
        return create_merged(file_new=file_new, file_old=file_old, file_merged=file_merged, mode=mode)

    results = [_create_merged(file_new=annotation_file, file_old=basic_level_file, file_merged=output_file, mode=mode)
               for annotation_file, basic_level_file, output_file
               in zip(annotation_files, basic_level_files, output_files)]

    # Output merging log to a csv file
    columns = ['duplicates_in_old_file', 'words_were_edited', 'words_were_added']
    results_df = pd.DataFrame(columns=columns,
                              data=results)
    results_df['exported_annotations_path'] = [annotation_file.absolute() for annotation_file in annotation_files]
    log = Path(f'merging_{mode}_log.csv')
    results_df.to_csv(path_or_buf=log, index=False)

    # Print numbers of files with duplicates, edited words and edited words
    duplicate_count, edited_count, added_count = results_df[columns].sum()
    print(f'There were:\n'
          f'{duplicate_count} basic level files with duplicate annotation ids.\n',
          f'{edited_count} merged files with words that have been edited.\n'
          f'{added_count} merged files with new words.\n\n'
          f'For details, see {log.absolute()}')


def _with_basic_level_folder(working_folder: Path, modality):
    """
    Returns the name of the folder to output annotations merged with previous basic level data.
    This function exists to avoid hard-coding the folder name in the functions that refer to it.
    :param working_folder: the parent folder
    :param modality: Audio/Video
    :return:
    """
    _check_modality(modality)
    return working_folder / f'with_basic_level_{modality.lower()}_annotations'


def merge_all_annotations_with_basic_level(
        exported_audio_annotations_folder, exported_video_annotations_folder,
        working_folder, exported_suffix='_processed.csv'):
    """
    Runs merge_annotations_with_basic_level on both audio and video annotations and puts the results to csv files in
    subfolders of working_folder.
    :param exported_audio_annotations_folder: folder to look for exported audio annotations in
    :param exported_video_annotations_folder: folder to look for exported video annotations in
    :param working_folder: the parent folder of the two output folders.
    :param exported_suffix: see merge_annotations_with_basic_level docstring
    :return:
    """
    # Audio
    with_basic_level_audio_folder = _with_basic_level_folder(working_folder, AUDIO)
    merge_annotations_with_basic_level(exported_annotations_folder=exported_audio_annotations_folder,
                                       output_folder=with_basic_level_audio_folder,
                                       mode='audio', exported_suffix=exported_suffix)

    # Video
    with_basic_level_video_folder = _with_basic_level_folder(working_folder, VIDEO)
    merge_annotations_with_basic_level(exported_annotations_folder=exported_video_annotations_folder,
                                       output_folder=with_basic_level_video_folder,
                                       mode='video', exported_suffix=exported_suffix)

    return with_basic_level_audio_folder, with_basic_level_video_folder


def make_incomplete_basic_level_list(merged_folder: Path):
    """
    Looks through all the files in the folder with annotations merged with previous basic level data and counts the
    number of rows that have to be manually updated
    :param merged_folder:
    :return: a pandas dataframe with two columns: filename, fixme_count
    """
    all_fixmes_df = None
    for csv_file in merged_folder.glob('*.csv'):
        fixmes_df = pd.read_csv(csv_file)
        fixmes_df = fixmes_df[fixmes_df.basic_level == FIXME]
        fixmes_df['filename'] = str(csv_file)
        all_fixmes_df = pd.concat([all_fixmes_df, fixmes_df])
    return all_fixmes_df


def is_any_missing_basic_level_data(merged_folder: Path, list_path: Path):
    """
    Runs make_incomplete_basic_level_list, saves it to a file and return whether there were any missing basic levels.
    :param merged_folder:
    :param list_path: where to output a list of rows missing basic level data
    :return: whether there were any rows with missing basic level data
    """
    df = make_incomplete_basic_level_list(merged_folder=merged_folder)
    df.to_csv(list_path, index=False)
    return df.size > 0


def check_all_basic_level_for_missing(merged_folder_audio, merged_folder_video, working_folder,
                                      raise_error_if_any_missing=True):
    """
    Runs is_any_missing_basic_level_data on both the audio and video folder with annotations merged with existing basic
    level data.
    :param merged_folder_audio:
    :param merged_folder_video:
    :param working_folder: the folder where list of missing basic level data will be saved if any
    :param raise_error_if_any_missing: should an error be raise if there are any missing?
    :return: were there any rows with missing basic levels?
    """
    missing_audio_basic_level_path = working_folder / 'missing_basic_level_audio.csv'
    is_missing_audio = is_any_missing_basic_level_data(merged_folder=merged_folder_audio,
                                                       list_path=missing_audio_basic_level_path)

    missing_video_basic_level_path = working_folder / 'missing_basic_level_video.csv'
    is_missing_video = is_any_missing_basic_level_data(merged_folder=merged_folder_video,
                                                       list_path=missing_video_basic_level_path)

    anything_missing = is_missing_audio or is_missing_video
    if anything_missing:
        if raise_error_if_any_missing:
            raise Exception('Some rows have missing basic level data. For details, see:\n'
                            f'{missing_audio_basic_level_path}\n'
                            f'{missing_video_basic_level_path}\n')
        else:
            return True
    else:
        return False


def export_all_annotations_to_csv(working_folder=None, ignore_audio_annotation_problems=False):
    """
    Exports audio and video annotations to csv files in subfolders of working_folder.
    :param working_folder: the parent folder of the output folders
    :param ignore_audio_annotation_problems: if False, will raise an exception if there were some problems when
    exporting audio annotations
    :return: tuple of paths to exported audio and video annotations respectively
    """
    working_folder = working_folder or Path('.')

    # Video annotations
    exported_video_annotations_folder = working_folder / 'exported_video_annotations'
    export_all_opfs_to_csv(exported_video_annotations_folder)

    # Audio annotations
    exported_audio_annotations_folder = working_folder / 'exported_audio_annotations'
    log = export_all_chas_to_csv(exported_audio_annotations_folder)
    if log and not ignore_audio_annotation_problems:
        raise Exception('There were problems during the export of audio annotations.'
                        ' See the following file for details:\n'
                        f'{log.absolute()}')

    return exported_audio_annotations_folder, exported_video_annotations_folder


def make_updated_basic_level_files(working_folder=None, ignore_audio_annotation_problems=False):
    """
    Creates updated versions of individual basic level files:
     - exports all annotations from cha and opf files, checks for exporting errors,
     - uses annotids to find basic level data in the current basic level files, mark rows where new one should be added.
    """
    working_folder = working_folder or Path('.')
    ensure_folder_exists_and_empty(working_folder)

    # Export
    exported_audio, exported_video = export_all_annotations_to_csv(
        working_folder=working_folder,
        ignore_audio_annotation_problems=ignore_audio_annotation_problems)

    # Merge with current basic level data
    merge_all_annotations_with_basic_level(
        exported_audio_annotations_folder=exported_audio,
        exported_video_annotations_folder=exported_video,
        working_folder=working_folder
    )

    print('\nThe annotations have been exported and merged with existing basic level data.\n'
          'Use scatter_updated_basic_level_files to check for basic levels that need updating amd move them to \n'
          'SubjectFiles.')


def scatter_updated_basic_level_files(working_folder=None):
    """
    Checks for missing basic level data in updated sparse_code csv files.
    If there are none, copies the files to their place on PN-OPUS, making a backup there first.
    :return:
    """
    working_folder = working_folder or Path('.')
    merged_folders = {modality: _with_basic_level_folder(working_folder, modality) for modality in (AUDIO, VIDEO)}

    anything_missing = check_all_basic_level_for_missing(
        merged_folder_audio=merged_folders[AUDIO],
        merged_folder_video=merged_folders[VIDEO],
        working_folder=working_folder,
        raise_error_if_any_missing=False)

    if anything_missing:
        print('\n'.join([
            'There were rows with missing basic level data. Check the "missing_basic_level_*.csv" files for a list of '
            'the rows.',
            '',
            '- Update the corresponding rows in the individual sparse_code csvs in the following folders:',
            f'  {merged_folders[AUDIO]}',
            f'  {merged_folders[VIDEO]}',
            '\n',
            '- Run scatter_updated_basic_level_files again.'
        ]))
        return

    for modality in (AUDIO, VIDEO):
        copy_all_basic_level_files_to_subject_files(
            updated_basic_level_folder=merged_folders[modality], modality=modality, backup=True)


def make_updated_all_basic_level_here():
    """
    Gathers all basic level files, checks for some errors, and - if there were no errors - writes four files:
    - all_basiclevel.csv
    - all_basiclevel.feather
    - all_basiclevel_NA.csv
    - all_basiclevel_NA.feather
    The files differ in whether they contain rows that have NA as the basic level and their format (csv/feather).
    The files are created in the current working directory. If the files already exist, they will be deleted first. This
    is done so there is a difference between:
    - everything went well, but there are no changes (files are the same, git status sees no changes) and
    - something went wrong (files are missing, git status will show that much).
    :return: None
    """
    # Delete current files
    output_paths = [Path(f'all_basiclevel{suffix}{extension}')
                    for suffix, extension in product(('', '_NA'), ('.csv', '.feather'))]
    for output_path in output_paths:
        try:
            output_path.unlink()
        except FileNotFoundError:
            pass

    # Gather all individual basic level files
    df_with_na = gather_all_basic_level_annotations(keep_basic_level_na=True)

    # Check for errors
    errors_df = check_for_errors(df_with_na)
    if errors_df:
        errors_file = 'errors.csv'
        logging.warning(f'The were errors found, the corresponding rows are in "{errors_file}".')
        errors_df.to_csv(errors_file)
        return

    # Write the four files
    def _write_to_csv_and_feather(all_basic_level_df, output_stem):
        write_all_basic_level_to_csv(all_basic_level_df=all_basic_level_df,
                                     csv_path=output_stem.with_suffix('.csv'))
        write_all_basic_level_to_feather(all_basic_level_df=all_basic_level_df,
                                         feather_path=output_stem.with_suffix('.feather'))
    # Without NAs
    output_stem_without_na = Path('all_basiclevel')
    df_without_na = df_with_na[~df_with_na.basic_level.isna()].reset_index(drop=True)
    # The feather version contains categorical information, so we need to remove categories that no longer exist - as if
    # the categories were set after removing NAs.
    for column_name in df_without_na.columns:
        if df_without_na[column_name].dtype.name == 'category':
            df_without_na[column_name] = df_without_na[column_name].cat.remove_unused_categories()
    _write_to_csv_and_feather(df_without_na, output_stem_without_na)

    # With NAs
    output_stem_with_na = output_stem_without_na.with_name(output_stem_without_na.name + '_NA')
    _write_to_csv_and_feather(df_with_na, output_stem_with_na)

    # Check that the output files have been created
    for output_path in output_paths:
        assert output_path.exists()
    print(f'all_basiclevel has been created and checked for errors, files have been written to.')


def calculate_listen_time_stats_for_cha_file(cha_path):
    """
    Runs listened_time.listen_time_stats_for_report on a single file accounting for files with four subregions.
    :param cha_path: path to the clan file
    :return: see listened_time.listen_time_stats_for_report
    """
    subregion_count = _get_subregion_count(**_parse_out_child_and_month(cha_path))
    clan_file_text = Path(cha_path).read_text()
    return listen_time_stats_for_report(clan_file_text=clan_file_text, subregion_count=subregion_count)


def preprocess_region_info(cha_path):
    """
    Extract enough info about the regions from a chat file to calculate listen time stats.
    This function does a half of what calculate_listen_time_stats_for_cha_file does.
    :param cha_path: path to the clan file
    :return: see `_preprocess_region_info`
    """
    subregion_count = _get_subregion_count(**_parse_out_child_and_month(cha_path))
    clan_file_text = Path(cha_path).read_text()
    return _preprocess_region_info(clan_file_text=clan_file_text, subregion_count=subregion_count)


def calculate_listen_time_stats_for_all_cha_files():
    """
    Runs calculate_listen_time_stats_for_cha_file on all cha files.
    :return: a pandas DataFrame with the calculated states and an additional column 'filename'
    """
    cha_paths = get_all_cha_paths()
    stats = [calculate_listen_time_stats_for_cha_file(cha_path) for cha_path in cha_paths]

    # Check uniqueness of the keys in all returned dicts
    [keys] = {tuple(stats_.keys()) for stats_ in stats}
    # Check that the keys are what we expect them to be
    expected_keys = ('num_makeup_region',
                     'num_extra_region',
                     'num_surplus_region',
                     'makeup_time',
                     'extra_time',
                     'surplus_time',
                     'subregion_time',
                     'num_subregion_with_annot',
                     'skip_silence_overlap_hour',
                     'skip_time',
                     'silence_time',
                     'silence_raw_hour',
                     'end_time',
                     'total_listen_time',
                     'positions',
                     'ranks',
                     'subregion_raw_hour',
                     'num_raw_subregion',
                     'annotation_counts_raw')
    assert keys == expected_keys

    # Combine into a dataframe and return
    return pd.DataFrame(
        index=pd.Index(data=[cha_path.name for cha_path in cha_paths], name='filename'),
        data=stats,
        columns=expected_keys).reset_index()



from zlib import adler32
from pathlib import Path
import contextlib
import os


def text_file_checksum(path: Path):
    """
    Returns the adler32 hash of the contents of a text file. Adler32 was used because it was faster than md5, there are
    certainly faster alternative I don't know about.
    :param path: path to the text file
    :return:
    """
    encoding = 'utf-8'
    # Decoding/encoding is done to stay invariant to different line endings.
    return adler32(path.read_text(encoding=encoding).encode(encoding=encoding))


class OutputExistsError(Exception):
    """
    Raised when a function crates output files and some of them already exist.
    """
    def __init__(self, paths):
        """
        Paths
        :param paths: the paths to the output files that already exist.
        """
        self.paths = paths
        message = 'Some of the output files already exist'
        if paths:
            message += ':\n\n' + '\n'.join((str(path.absolute()) for path in paths))
        super().__init__(message)


# copied from
@contextlib.contextmanager
def modified_environ(*remove, **update):
    """
    Temporarily updates the ``os.environ`` dictionary in-place.
    The ``os.environ`` dictionary is updated in-place so that the modification
    is sure to work in all situations.
    :param remove: Environment variables to remove.
    :param update: Dictionary of environment variables and values to add/update.
    """
    env = os.environ
    update = update or {}
    remove = remove or []

    # List of environment variables being updated or removed.
    stomped = (set(update.keys()) | set(remove)) & set(env.keys())
    # Environment variables and values to restore on exit.
    update_after = {k: env[k] for k in stomped}
    # Environment variables and values to remove on exit.
    remove_after = frozenset(k for k in update if k not in env)

    try:
        env.update(update)
        [env.pop(k, None) for k in remove]
        yield
    finally:
        env.update(update_after)
        [env.pop(k) for k in remove_after]


def df_to_list_of_tuples(df):
    return list(df.to_records(index=False))



from datetime import date
from pathlib import Path
from shutil import copy2

import pandas as pd

from .paths import _parse_out_child_and_month, get_basic_level_path, _check_modality, AUDIO, VIDEO
from .merge import read_annotations_csv


def backup_to_old_files(file_path: Path):
    """
    Move file at file_path to the folder "Old_Files" at the same level adding date in the format "_YYYY-MM-DD". The
    "Old_Files" will be created if necessary.
    throw a FileExistsError if there is an already existing backup with the same date.

    Note: a.tar.gz will be renamed to a.tar_2021-12-30.gz

    :param file_path: path to the file to be backed
    """
    assert file_path.is_file()

    date_string = date.today().isoformat()
    backup_path = file_path.parent / 'Old_Files' / f'{file_path.stem}_{date_string}{file_path.suffix}'

    backup_path.parent.mkdir(exist_ok=True)

    if backup_path.exists():
        raise FileExistsError('Can\'t back up\n'
                              f'\t{file_path.absolute()}\n'
                              '\tto\n'
                              f'\t{backup_path.absolute()}\n'
                              '\tbecause the second path already exists.')

    copy2(file_path, backup_path)


def sort_basic_level_df(df, modality):
    """
    Sorts a dataframe read from an individual basic level file (sparse_code csv)
    :param df: a pandas dataframe
    :param modality: Audio/Video
    :return: df sorted
    """
    _check_modality(modality)
    # TODO: sort all files by "ordinal" column once it is added to the cha export. Probably kill this function while at
    # it.
    if modality == VIDEO:
        return df.sort_values(by='labeled_object.ordinal')
    elif modality == AUDIO:
        # Sort by onset, offset, annotid. Onsets and offsets have to be extracted first.
        sorting_index = (pd.concat([df.timestamp.str.extract(r'(?P<onset>\d+)_(?P<offset>\d+)').astype(int),
                                    df.annotid],
                                   axis='columns')
                         .sort_values(by=['onset', 'offset', 'annotid'])
                         .index)
        return df.loc[sorting_index]


def copy_basic_level_to_subject_files(file_path: Path, modality, backup=True):
    """
    Copies the basic level file at file_path to the corresponding folder in the Seedlings folder. The correspondence is
    established based on the child and month number in the filename and the modality argument. The older version is
    backed up first.
    :param file_path: path to the newer basic level file
    :param modality: Auido/Video
    :return: None
    """
    _check_modality(modality)

    # Check that the file looks like an actual basic level file of the right modlaity
    # It is an existing csv file
    assert file_path.exists() and file_path.is_file() and file_path.name.endswith('.csv')
    # With 'basic_level' in the column definitions
    with file_path.open() as f:
        assert 'basic_level' in f.readline()
    # And modality in its name (something like '01_06_audio_sparse_code.csv')
    assert modality.lower() in file_path.name.lower()

    # Sort the rows in the source file and overwrite it
    sort_basic_level_df(df=read_annotations_csv(file_path), modality=modality).to_csv(file_path, index=False)

    # Backup the current version
    basic_level_path = get_basic_level_path(**_parse_out_child_and_month(file_path), modality=modality)
    if backup:
        backup_to_old_files(basic_level_path)

    # Copy the new version
    copy2(file_path, basic_level_path)


def copy_all_basic_level_files_to_subject_files(updated_basic_level_folder: Path, modality, backup=True):
    """
    Runs copy_basic_level_to_subject_files on all csv files in a folder.
    :param updated_basic_level_folder: folder with the basic level files
    :param modality: Audio/Video
    :param backup: should csv files be backed up to "Old_Files" first?
    successfully copied would not be attempted to copy again leading to error because a backup file already exists.
    :return: None
    """
    for basic_level_path in updated_basic_level_folder.glob('*.csv'):
        copy_basic_level_to_subject_files(file_path=basic_level_path, modality=modality, backup=backup)



import pandas as pd

from blabpy.seedlings.gather import check_for_errors


def test_check_for_errors():
    columns = ['annotid', 'utterance_type', 'object_present']
    inputs = [
        pd.DataFrame(columns=columns, data=data)
        for data in (
            [('0xa', 'd', 'y'),
             ('0xa', 'r', 'n'),  # same annotation id
             ('0xb', 'y', 'n'),  # invalid utterance type code
             ('0xc', 'd', 'Y')],  # invalid object present code
            [('0xa', 'd', 'y'),
             ('0xa', 'r', 'n')],  # same annotation id
            [('0xa', 'd', 'y'),
             ('0xb', 'y', 'n')],  # invalid utterance type code
            [('0xa', 'd', 'Y'),  # invalid object present code
             ('0xb', 'r', 'n')],
            [('0xa', 'd', 'y'),  # all fine
             ('0xb', 'd', 'y'),
             ('0xc', 'd', 'y'),
             ('0xd', 'r', 'n')]
        )]
    outputs = [
        pd.DataFrame(columns=(['error_type'] + columns), data =data)
        for data in (
            [['duplicate annotation id', '0xa', 'd', 'y'],
             ['duplicate annotation id', '0xa', 'r', 'n'],
             ['invalid utterance type code', '0xb', 'y', 'n'],
             ['invalid object present code', '0xc', 'd', 'Y']],
            [['duplicate annotation id', '0xa', 'd', 'y'],
             ['duplicate annotation id', '0xa', 'r', 'n']],
            [['invalid utterance type code', '0xb', 'y', 'n']],
            [['invalid object present code', '0xa', 'd', 'Y']]
        )]

    for input_, output in zip(inputs[:-1], outputs):
        assert check_for_errors(input_).equals(output)

    assert check_for_errors(inputs[-1]) is None



from pathlib import Path
from shutil import copy2

from blabpy.vihi import pipeline
from blabpy.vihi.pipeline import add_intervals_for_annotation
from blabpy.vihi import paths as vihi_paths


from blabpy.vihi.intervals.tests.test_intervals import _read_intervals_with_fake_metric, _read_sub_recordings, \
    _read_intervals, _read_test_vtc_data, _get_test_eaf_path, _get_selected_regions_path, \
    _get_expected_with_extra_eaf_path

TEST_FULL_RECORDING_ID = 'TEST_123_290'


def _test_region_output_files(tmpdir):
    return {
        'eaf': Path(tmpdir / f'{TEST_FULL_RECORDING_ID}.eaf'),
        'pfsx': Path(tmpdir / f'{TEST_FULL_RECORDING_ID}.pfsx'),
        'csv': Path(tmpdir / f'{TEST_FULL_RECORDING_ID}_selected-regions.csv')}


def test_add_intervals_for_annotation(monkeypatch, tmpdir):
    # Monkeypatch functions that look into the VIHI folder on pn-opus.

    # Since we are using a full recording id that can't be in the VIHI folder, the worst thing that can happen,
    # if we forget to patch some functions or `add_intervals_for_annotation` starts using different functions, is the
    # test erroring out before even getting to any assertions.

    region_output_files = _test_region_output_files(tmpdir)

    def __test_region_output_files(*args, **kwargs):
        # The output files need to go to the tempdir
        return region_output_files

    # TODO: that's way too much monkeypatching, break down add_intervals_for_annotation into parts
    monkeypatch.setattr(pipeline, '_region_output_files', __test_region_output_files)
    monkeypatch.setattr(pipeline, 'get_eaf_path_from_full_recording_id', _get_test_eaf_path)
    monkeypatch.setattr(pipeline, 'get_vtc_data', _read_test_vtc_data)
    monkeypatch.setattr(pipeline, 'make_intervals', lambda *args, **kwargs: _read_intervals())
    monkeypatch.setattr(pipeline, 'gather_recordings', lambda *args, **kwargs: _read_sub_recordings())
    monkeypatch.setattr(pipeline, 'add_metric', lambda *args, **kwargs: _read_intervals_with_fake_metric())
    monkeypatch.setattr(vihi_paths, 'POPULATIONS', ['TEST'])
    monkeypatch.setattr(pipeline, 'INTERVALS_FOR_ANNOTATION_COUNT', 2)
    monkeypatch.setattr(pipeline, 'INTERVALS_EXTRA_COUNT', 1)

    # Copy "selected_regions.csv" to
    copy2(_get_selected_regions_path(), region_output_files['csv'])

    # Add intervals
    add_intervals_for_annotation(TEST_FULL_RECORDING_ID)

    # Check the output file
    eaf_path = region_output_files['eaf']
    expected_eaf_path = _get_expected_with_extra_eaf_path()
    assert eaf_path.read_text() == expected_eaf_path.read_text()



"""
Functions to facilitate working with LENA's .its files.
"""
from pathlib import Path
from xml.etree.ElementTree import ElementTree, XMLParser

import pandas as pd


class ItsNoTimeZoneInfo(Exception):
    pass


class Its(object):
    """
    Container for the parsed xml from a single .its file.
    """
    def __init__(self, xml_parsed: ElementTree, forced_timezone=None):
        """
        :param xml_parsed: Parsed xml from an its file.
        :param forced_timezone: if forced_timezone is not None, the timezone in the file will be ignored and this
        one will be used instead. Useful for files with incorrect timezone information or without any information at
        all. The format is:
        dict(seconds_offset=<offset from UTC in seconds>, uses_dst=<True if DST was used, False otherwise>)
        """
        self.xml: ElementTree = xml_parsed
        if forced_timezone is not None:
            self._check_timezone(forced_timezone)
        self.forced_timezone = forced_timezone

    @staticmethod
    def _check_timezone(timezone_dict):
        assert type(timezone_dict) is dict
        assert 'seconds_offset' in timezone_dict and type(timezone_dict['seconds_offset']) is int
        assert 'uses_dst' in timezone_dict and type(timezone_dict['uses_dst']) is bool

    @classmethod
    def from_string(cls, its_contents: str, forced_timezone=None):
        """
        Parses a string with .its file contents.
        :param its_contents: single string containing contents of an its file.
        :param forced_timezone: see __init__'s docstring
        :return: Its object with its_contents parsed
        """
        parser = XMLParser()
        parser.feed(its_contents)
        xml_parsed = parser.close()

        return Its(xml_parsed=xml_parsed, forced_timezone=forced_timezone)

    @classmethod
    def from_path(cls, its_path, forced_timezone=None):
        """
        Reads and parses an its file.
        :param its_path: path to an its file as a string or as a pathlib.Path object
        :param forced_timezone: see __init__'s docstring
        :return: Its object
        """
        its_path = Path(its_path)
        return cls.from_string(its_contents=its_path.read_text(), forced_timezone=forced_timezone)

    def get_timezone_info(self):
        if self.forced_timezone is not None:
            return self.forced_timezone

        timezone_xml_path = './ProcessingUnit/UPL_Header/TransferredUPL/RecordingInformation/Audio/TimeZone'
        timezone_element = self.xml.find(timezone_xml_path)
        if timezone_element is None:
            raise ItsNoTimeZoneInfo('I wasn\'t able to locate timezone info in this .its file')

        timezone_info = dict(timezone_element.items())
        return dict(seconds_offset=timezone_info['StandardSecondsOffset'], uses_dst=timezone_info['UsesDST'])

    def convert_utc_to_local(self, clock_time: pd.Series):
        """
        Convert utc timestams to local timestamps.
        :param clock_time: pd.Series of strings from the .its object
        :return: pd.Series of timezone-naive datetime type.
        """
        # Parse
        datetimes = (pd.to_datetime(clock_time, format='%Y-%m-%dT%H:%M:%SZ', utc=True)  # parse date
                     .dt.tz_convert(None))  # remove timezone information

        # Apply the timezone difference
        # Note: I didn't figure out how to use 'GMT-05:00' that is in timezone_info['ZoneNameShort']
        timezone_info = self.get_timezone_info()
        datetimes += pd.Timedelta(seconds=int(timezone_info['seconds_offset']))

        # Add an hour if there is daylight savings time used
        if timezone_info['uses_dst']:
            datetimes += pd.Timedelta(hours=1)

        return datetimes

    @staticmethod
    def convert_iso_duration_to_ms(iso_time: pd.Series):
        """
        Converts duration from iso format to the number of milliseconds.
        :param iso_time: pd.Series of ISO duration values
        :return: pd.Series of ints
        """
        return (pd.to_timedelta(iso_time)  # parse ISO-formatted times
                .dt.total_seconds()  # convert to seconds as real numbers
                .multiply(1000)  # convert from s to ms
                .astype(int))  # finally, convert to integers

    def gather_recordings(self):
        """
        Finds all sub-recordings in the its file.

        This works very similarly to rlena::gather_recordings.

        :return: a pd.DataFrame with the following columns:
        - recording_start, recording_end - datetime columns,
        - recording_start_wav - integer column of when recording started in ms within the wav file.
        """
        # Get raw recordings info
        recordings_xml_path = './ProcessingUnit/Recording'
        recordings_elements = self.xml.findall(path=recordings_xml_path)
        # Each element's items are (key, value) tuples which we'll convert to a dict
        recordings_list = [dict(element.items()) for element in recordings_elements]
        recordings = pd.DataFrame.from_records(recordings_list)
        assert set(recordings.columns) == {'num', 'startClockTime', 'endClockTime', 'startTime', 'endTime'}
        # num - order number
        # *ClockTime - UTC time when recordings started/ended
        # *Time - ISO-formatted duration of time from the start of the wav to the start/end of the recording

        # Let's convert *ClockTime to local time and startTime to milliseconds
        return recordings.assign(
            recording_start=lambda df: self.convert_utc_to_local(df.startClockTime),
            recording_end=lambda df: self.convert_utc_to_local(df.endClockTime),
            recording_start_wav=lambda df: self.convert_iso_duration_to_ms(df.startTime)
        )[['recording_start', 'recording_end', 'recording_start_wav']]



# Tests blabpy.vihi.intervals.cli.cli_batch_create_files_with_random_regions
# I wasn't able to write a proper test so this was run by hand :-(
# Mostly, I was scared to mess up some actual VIHI files.

from pathlib import Path

from blabpy.paths import PN_OPUS, get_pn_opus_path
from blabpy.vihi.paths import get_lena_recording_path, parse_full_recording_id
from blabpy.vihi.intervals.cli import cli_batch_create_files_with_random_regions
from blabpy.utils import modified_environ

import pandas as pd

info_spreadsheet = \
    pd.DataFrame(columns='id,age,length_of_recording'.split(','),
                 data=('VI_666_924,30,960'.split(','),
                       'VI_777_234,12,360'.split(','),
                       'VI_888_098,17,640'.split(',')))

zhenya_pn_opus_mock = Path('/Volumes/pn-opus/VIHI/WorkingFiles/zhenya')

pn_opus_path = get_pn_opus_path()
for row in info_spreadsheet.itertuples():
    path = get_lena_recording_path(**parse_full_recording_id(row.id))
    zhenya_pn_opus_mock.joinpath(path.relative_to(pn_opus_path)).mkdir(exist_ok=True, parents=True)

# Works fine
info_spreadsheet_path = zhenya_pn_opus_mock / 'info_spreadsheet.csv'
info_spreadsheet.iloc[:2].to_csv(info_spreadsheet_path)
with modified_environ(**{PN_OPUS: str(zhenya_pn_opus_mock)}):
    args = [str(info_spreadsheet_path)]
    cli_batch_create_files_with_random_regions(args=args)

# Throws an error
info_spreadsheet.iloc[1:].to_csv(info_spreadsheet_path)
with modified_environ(**{PN_OPUS: str(zhenya_pn_opus_mock)}):
    args = [str(info_spreadsheet_path)]
    cli_batch_create_files_with_random_regions(args=args)



import pandas as pd
import pytest

from blabpy.seedlings.listened_time import _set_difference_of_intervals, \
    _remove_interval_from_regions, _remove_silences_and_skips, _total_time_and_count_per_region_type, \
    _remove_subregions, \
    _overlaps_with_interval, RegionType, _assert_no_overlaps, _extract_region_info

test_regions_df = pd.DataFrame(
    columns=['region_type', 'start', 'end', 'position'],
    data=[['surplus', 3600710, 6301520, 1],
          ['subregion', 6300000, 9900000, 1],
          ['extra', 9900230, 10200470, 1],
          ['surplus', 10516250, 10800010, 2],
          ['silence', 11183120, 15599480, 1],
          ['subregion', 15300000, 18900000, 2],
          ['surplus', 15793420, 18900930, 3],
          ['subregion', 18900000, 22500000, 3],
          ['silence', 26255960, 26733920, 2],
          ['silence', 27019130, 27504220, 3],
          ['subregion', 27600000, 31200000, 4],
          ['subregion', 33600000, 37200000, 5],
          ['silence', 44867230, 57599990, 4]])


def test__set_difference_of_intervals():
    # Invalid inputs should raise an error
    with pytest.raises(AssertionError):
        _set_difference_of_intervals((0, 1), (3, 2))
    with pytest.raises(AssertionError):
        _set_difference_of_intervals((1, 0), (2, 3))
    with pytest.raises(AssertionError):
        _set_difference_of_intervals((1, 0), (3, 2))

    # All inputs should be integers
    for minuend, subtrahend in ((('1', 12), (2, 21)),
                                ((1, '12'), (2, 21)),
                                ((1, 12), ('2', 21)),
                                ((1, 12), (2, '21'))):
        with pytest.raises(AssertionError):
            _set_difference_of_intervals(minuend, subtrahend)

    # Check that the output is correct
    minuend = (-1, 1)
    subtrahends, correct_differences = zip(*(
        ((-3, -2), [(-1, 1)]),
        ((-3, -1), [(-1, 1)]),
        ((-3, 0), [(0, 1)]),
        ((-3, 1), []),
        ((-3, 2), []),
        ((-1, 0), [(0, 1)]),
        ((-1, 1), []),
        ((-1, 2), []),
        ((0, 0.5), [(-1, 0), (0.5, 1)]),
        ((0, 1), [(-1, 0)]),
        ((0, 2), [(-1, 0)]),
        ((1, 2), [(-1, 1)]),
        ((2, 3), [(-1, 1)])
    ))
    for subtrahend, correct_difference in zip(subtrahends, correct_differences):
        assert _set_difference_of_intervals(minuend, subtrahend) == correct_difference


def test__remove_interval_from_regions():
    regions_list = [pd.DataFrame.from_dict({'start': [0, 4, 8],
                                            'end': [2, 6, 10],
                                            'arbitrary_column': ['a', 'b', 'c']}),
                    pd.DataFrame.from_dict({'start': [0],
                                            'end': [10],
                                            'arbitrary_column': ['a']})]
    # Interval to be removed
    start, end = 1, 9

    correct_results = [pd.DataFrame.from_dict({'start': [0, 9],
                                               'end': [1, 10],
                                               'arbitrary_column': ['a', 'c']}),
                       pd.DataFrame.from_dict({'start': [0, 9],
                                               'end': [1, 10],
                                               'arbitrary_column': ['a', 'a']})]

    for regions, correct_result in zip(regions_list, correct_results):
        regions_copy = regions.copy()
        assert _remove_interval_from_regions(regions=regions_copy, start=start, end=end).equals(correct_result)
        # Check that the input has not been modified
        assert regions.equals(regions_copy)


def test__remove_silences_and_skips():
    regions = pd.DataFrame(
        columns=['region_type', 'start', 'end', 'position'],
        data=[('subregion', 1, 10, 1),
              ('subregion', 11, 20, 2),
              ('subregion', 22, 30, 3),
              ('silence', 5, 21, 1),
              ('skip', 24, 27, 1),
              ('skip', 32, 37, 2)])

    correct_result = pd.DataFrame(**{
        'columns': ['region_type', 'start', 'end', 'position'],
        'data': [['skip', 24, 27, 1],
                 ['skip', 32, 37, 2],
                 ['silence', 5, 21, 1],
                 ['subregion', 1, 5, 1],
                 ['subregion', 22, 24, 3],
                 ['subregion', 27, 30, 3]]})

    assert _remove_silences_and_skips(regions).equals(correct_result)


def test__total_time_per_region_type():
    correct_result = pd.DataFrame(
        columns=['region_type', 'total_time', 'count'],
        data=[['extra', 300240, 1],
              ['silence', 18112170, 4],
              ['subregion', 18000000, 5],
              ['surplus', 6092080, 3]]).set_index('region_type')
    assert _total_time_and_count_per_region_type(test_regions_df).equals(correct_result)


def test__remove_subregions():
    correct_result = pd.DataFrame(**{
        'columns': ['region_type', 'start', 'end', 'position'],
        'data': [['surplus', 3600710, 6301520, 1],
                 ['extra', 9900230, 10200470, 1],
                 ['surplus', 10516250, 10800010, 2],
                 ['silence', 11183120, 15599480, 1],
                 ['surplus', 15793420, 18900930, 3],
                 ['silence', 26255960, 26733920, 2],
                 ['silence', 27019130, 27504220, 3],
                 ['subregion', 27600000, 31200000, 4],
                 ['subregion', 33600000, 37200000, 5],
                 ['silence', 44867230, 57599990, 4]]})
    result = _remove_subregions(test_regions_df,
                                condition_function=_overlaps_with_interval,
                                other_region_types=[RegionType.SURPLUS])
    assert result.equals(correct_result)


def test__assert_no_overlaps():
    overlapping = pd.DataFrame(columns=['start', 'end'], data=((0, 2), (1, 3)))
    non_overlapping = pd.DataFrame(columns=['start', 'end'], data=((0, 1), (2, 3)))
    with pytest.raises(AssertionError):
        _assert_no_overlaps(overlapping)
    _assert_no_overlaps(non_overlapping)


def test__extract_region_info():
    clan_file_text = '\n'.join([
        '*OLN:\t0 . \x15653650_654740\x15',
        '%xcom:\tbegin skip, eating',
        '*NOF:\t0 . \x15699220_701620\x15',
        '%xcom:\tend skip, eating over',
        'whatever'
    ])
    expected_region_boundaries_df = pd.DataFrame.from_dict({'region_type': {0: 'skip'},
                                                            'start': {0: 654740},
                                                            'end': {0: 701620},
                                                            'position': {0: 1}})

    region_boundaries_df, _, _ = _extract_region_info(clan_file_text=clan_file_text, subregion_count=0)
    assert region_boundaries_df.equals(expected_region_boundaries_df)



from importlib.resources import path

import blabpy.vihi.intervals.etf_templates as etf_templates


def _get_etf_template(filename):
    """
    I don't know why but the path function returns a context manager, not a path.
    :param filename: a template filename
    :return:
    """
    with path(etf_templates, filename) as p:
        return p


basic_00_07 = _get_etf_template('ACLEW-basic-template_00-07mo.etf')
basic_08_18 = _get_etf_template('ACLEW-basic-template_08-18mo.etf')
basic_19_36 = _get_etf_template('ACLEW-basic-template_19-36mo.etf')
basic_00_07_pfsx = _get_etf_template('ACLEW-basic-template_00-07mo.pfsx')
basic_08_18_pfsx = _get_etf_template('ACLEW-basic-template_08-18mo.pfsx')
basic_19_36_pfsx = _get_etf_template('ACLEW-basic-template_19-36mo.pfsx')


def choose_template(*, age_in_months):
    if 0 <= age_in_months <= 7:
        return basic_00_07, basic_00_07_pfsx
    elif 8 <= age_in_months <= 18:
        return basic_08_18, basic_08_18_pfsx
    elif 19 <= age_in_months <= 36:
        return basic_19_36, basic_19_36_pfsx



import csv
import re
import os
from zipfile import ZipFile, ZIP_DEFLATED
import tempfile
from pathlib import Path

import pandas as pd


# This is not exactly correct. datavyu uses milliseconds and this uses microseconds adding three extra zeros
DATETIME_FORMAT = '%H:%M:%S:%f'


class OPFFile(object):
    SKIP_PREFIXES = ('.DS_Store', '__MACOSX/')

    def __init__(self, path):
        self.path: Path = path
        self.loaded = False
        self.db = None
        self.other_components = None
        self.name_list = None
        self.load()

    def load(self):
        if self.path.is_dir():
            # List all the files skipping macos-specific hidden files
            name_list = [path.name for path in self.path.iterdir()
                         if not any(path.name.startswith(prefix) for prefix in self.SKIP_PREFIXES)]

            # Load the annotations
            assert 'db' in name_list, f'The folder at {self.path} does not contain a "db" file. Not an OPF folder?'
            db = self.path.joinpath('db').read_text(encoding='utf-8')

            # Other files
            other_components = {
                name: self.path.joinpath(name).read_bytes()  # for parity with the archived version (see below)
                for name in name_list
                if name != 'db'
            }

        elif self.path.suffix == '.opf':
            with ZipFile(self.path, 'r') as opf_zipped:
                # List all the files skipping macos-specific hidden files
                name_list = [fn for fn in opf_zipped.namelist()
                             if not any(fn.startswith(prefix) for prefix in self.SKIP_PREFIXES)]
                assert 'db' in name_list, f'The file at {self.path} does not contain "db". Not an OPF file?'

                # Load the annotations
                with opf_zipped.open('db', 'r') as db_zipped:
                    # ZipFile.open reads files in the binary mode
                    db = db_zipped.read().decode('utf-8')

                # Other files
                other_components = {
                    name: opf_zipped.open(name, 'r').read()
                    for name in name_list
                    if name != 'db'}

        else:
            raise ValueError('Can read OPF annotation from .opf files or unzipped folders, the input path is neither:'
                             + str(self.path))

        self.db = db
        self.name_list = name_list
        self.other_components = other_components
        self.loaded = True

    def read_in_editor(self):
        zf = ZipFile(self.path)
        tempdir = tempfile.mkdtemp()
        zf.extractall(tempdir)
        db_path = os.path.join(tempdir, 'db')
        os.system(f'open {db_path}')

    def write(self, path: Path = None, overwrite_original=False, unzipped=False):
        if not path and not overwrite_original:
            raise ValueError('You haven\'t specified the path to write the files. If you want to overwrite the original'
                             ' file, set overwrite_original to True')

        if path and overwrite_original:
            raise ValueError('You\'ve specified the path and set overwrite_original to True - you can only do one of '
                             'those')

        path = path or self.path

        if unzipped:
            if path.exists() and not path.is_dir():
                raise ValueError('Unzipped is set to True but supplied path is not a directory.')
            else:
                path.mkdir(parents=True, exist_ok=True)
                self._write_to_dir(folder_path=path)
        else:
            if not path.name.endswith('.opf'):
                raise ValueError('Supplied path does not end with .opf as expected')
            else:
                self._write_to_opf(path)

    def _write_to_dir(self, folder_path):
        for filename in self.name_list:
            filepath = folder_path / filename
            if filename == 'db':
                with filepath.open('w') as f:
                    f.write(''.join(self.db))
            else:
                with filepath.open('wb') as f:
                    f.write(self.other_components[filename])

    def _write_to_opf(self, path):
        with ZipFile(path, mode='w', compression=ZIP_DEFLATED) as opf_zipped:
            for filename in self.name_list:
                if filename == 'db':
                    opf_zipped.writestr('db', self.db)
                else:
                    opf_zipped.writestr(filename, self.other_components[filename])

    def update_annotations(self, opf_df: 'OPFDataFrame'):
        """
        Updates the annotations based on an OPFDataFrame object with updated data.
        Preserves presence/absence of newline at the end of file to avoid annoying diffs.
        :param opf_df:
        :return:
        """
        newline_at_end_of_file = self.db.endswith('\n')
        self.db = str(opf_df)
        if newline_at_end_of_file:
            self.db = self.db.rstrip('\n') + '\n'


class OPFDataFrame(object):
    def __init__(self, opf_file: OPFFile):
        self.opf_file = opf_file
        self.prefix = None
        self.column_definitions = None
        self.df = self._opf_to_pandas_df()
        # assert self._can_be_reversed()

    def _opf_to_pandas_df(self):
        db_lines = self.opf_file.db.splitlines()
        # Sometimes the last line is empty - delete it
        if not db_lines[-1]:
            del db_lines[-1]
        self.prefix = db_lines[0]
        self.column_definitions = db_lines[1]

        # Extract field names
        # There is a single datavyu column "labeled_object" defined in the second line of "db".
        # The format of this line is <column-definition>-<field_definitions>
        field_definitions = self.column_definitions.split('-')[1]
        # Field definitions are comma-separated, each definition has the following format: <field_name>|<field_type>
        field_names = [field_definition.split('|')[0] for field_definition in field_definitions.split(',')]
        # The first two columns contain timestamps
        field_names = ['time_start', 'time_end'] + field_names

        # Extract values
        # Each data row in db is in this format: <time_start>,<time_end>,(<field1>,...,<fieldN>)
        def row_to_values(row):
            values = row.split(',', maxsplit=2)
            # Commas within filed values are escaped by a backslash - we don't want to split on those
            values = values[:2] + re.split(r'(?<!\\),', values[2].strip('()'))
            # If, for some reason, a row is missing commas (not just values!), pad it with empty fields
            values = values + [''] * (len(field_names) - len(values))
            return values

        data = list(map(row_to_values, db_lines[2:]))

        # Bind
        df = pd.DataFrame(columns=field_names, data=data)

        return df

    def __str__(self):
        """
        Converts back to text format
        :return: str
        """
        # Reformat data into single column of the format <time>,<time>,(<col1>,...,<col2>)
        df = self.df
        time_columns = ('time_start', 'time_end')
        other_columns = ~df.columns.isin(time_columns)
        data = (pd.concat([
            # Time columns
            df.loc[:, time_columns],
            # All the other columns put together in parentheses and separated by commas
            df.loc[:, other_columns]
              .astype(str)
              .agg(lambda values: f'({",".join(values)})', axis=1)
            ], axis=1)
            .agg(','.join, axis=1)
        )

        return '\n'.join([self.prefix,
                          self.column_definitions,
                          *data.to_list()])

    def can_be_reversed(self):
        """
        Can we reconstruct the db in the original file up to an empty line at the end?
        :return:
        """
        return str(self) == self.opf_file.db.rstrip()

    @staticmethod
    def time_column_to_milliseconds(time_str: pd.Series):
        dt = pd.to_datetime(time_str, format=DATETIME_FORMAT)
        return ((dt.dt.hour * 60 + dt.dt.minute) * 60 + dt.dt.second) * 1000 + dt.dt.microsecond / 1000


def export_opf_to_csv(opf_path, csv_path):
    """
    Emulate datavyu export
    :param opf_path: Path to the opf
    :param csv_path: Path to the output csv
    :return:
    """
    # Load the data
    df = OPFDataFrame(OPFFile(opf_path)).df

    # Make sure the index is 1, 2, 3 and make it a column
    df = df.reset_index(drop=True).reset_index()
    df['index'] += 1

    # Rename columns
    df.rename(columns={
        'index': 'ordinal',
        'time_start': 'onset',
        'time_end': 'offset'
    }, inplace=True)

    # Convert time to milliseconds
    df['onset'] = OPFDataFrame.time_column_to_milliseconds(df.onset).astype(int)
    df['offset'] = OPFDataFrame.time_column_to_milliseconds(df.offset).astype(int)

    # Prepend "labeled_object." to column names
    df.columns = 'labeled_object.' + df.columns

    # Write the output manually. Specifics of the datavyu export function (adding a comma to the end of each line) make
    # it harder to use df.to_csv directly.
    with csv_path.open('w') as f:
        columns, *rows = df.to_csv(index=False, quoting=csv.QUOTE_NONNUMERIC).splitlines()
        # There is an extra comma in the datavyu export for some reason.
        suffix = ',\n'

        # Write the column names which are not quoted in datavyu output
        f.write(columns.replace('"', '') + suffix)

        # And then the lines, in which non-numeric data *are* quoted
        for row in rows:
            f.write(row + suffix)


from pathlib import Path
from random import seed

import pandas as pd
from pandas.testing import assert_frame_equal
import pytest
from pydub.generators import WhiteNoise

import blabpy.vihi.intervals.intervals as intervals_module
from blabpy.eaf import EafPlus
from blabpy.utils import OutputExistsError, text_file_checksum
from blabpy.vihi.intervals.intervals import calculate_energy_in_all_intervals, create_files_with_random_regions, \
    batch_create_files_with_random_regions, make_intervals, select_best_intervals, add_metric, \
    add_annotation_intervals_to_eaf, CONTEXT_BEFORE, CODE_REGION, CONTEXT_AFTER
from blabpy.vihi.paths import compose_full_recording_id
from blabpy.vtc import read_rttm

DATA_PATH = Path(__file__).parent / 'data'
# Intervals already in eaf before we add high-volubility ones
PRE_EXISTING_CODE_INTERVALS = [(353000, 473000), (1200000, 1320000)]


def test_calculate_energy_in_all_intervals():
    seed(24)
    noise = WhiteNoise().to_audio_segment(duration=200)
    intervals = pd.DataFrame.from_dict(dict(start=[0, 50, 150], end=[50, 150, 200]))

    # Without filtering
    energy = calculate_energy_in_all_intervals(intervals=intervals, audio=noise)
    expected_energy = pd.Series({0: 733.6411476029159, 1: 1469.4191753712091, 2: 728.1696980215235})
    assert energy.equals(expected_energy)

    # With filtering
    energy = calculate_energy_in_all_intervals(intervals=intervals, audio=noise, low_freq=300, high_freq=3000)
    expected_energy = pd.Series({0: 82.11954411321199, 1: 187.31203579565587, 2: 81.01795951661462})
    assert energy.equals(expected_energy)


def test_create_files_with_random_regions(monkeypatch, tmp_path):
    # Make sure the test doesn't affect real files
    full_recording_id = 'TD_666_222'
    recording_path = tmp_path / full_recording_id
    recording_path.mkdir()
    monkeypatch.setattr(intervals_module, 'get_lena_recording_path', lambda *args, **kwargs: recording_path)

    # Run the first time
    def run():
        create_files_with_random_regions(full_recording_id=full_recording_id, age=12, length_of_recording=360)

    run()

    # Check that the files have been created
    expected_filenames = ['TD_666_222_selected-regions.csv', 'TD_666_222.eaf', 'TD_666_222.pfsx']
    assert all(recording_path.joinpath(filename).exists for filename in expected_filenames)

    # TODO: check file contents too.

    # Trying to run again should raise an error
    with pytest.raises(OutputExistsError):
        run()


def test_batch_create_files_with_random_regions(monkeypatch, tmp_path):
    # Make sure pn-opus is not touched
    def get_lena_recording_path_(population, subject_id, recording_id):
        return tmp_path / compose_full_recording_id(population, subject_id, recording_id)
    monkeypatch.setattr(intervals_module, 'get_lena_recording_path', get_lena_recording_path_)

    # Prepare the recordings list
    info_spreadsheet_path_1 = tmp_path / 'info_spreadsheet.csv'
    info_spreadsheet_1 = pd.DataFrame(columns='id,age,length_of_recording'.split(','),
                                      data=('VI_666_924,30,960'.split(','),
                                            'VI_777_234,12,360'.split(',')))
    info_spreadsheet_1.to_csv(info_spreadsheet_path_1, index=False)

    # Create the recordings folders
    info_spreadsheet_1.id.apply(lambda full_recording_id: tmp_path.joinpath(full_recording_id).mkdir())

    # Run once
    batch_create_files_with_random_regions(info_spreadsheet_path_1, seed=7)
    
    # Compare the output files
    expected_file_checksums = [('VI_666_924/VI_666_924.eaf', 580209740),
                               ('VI_666_924/VI_666_924.pfsx', 1301328091),
                               ('VI_666_924/VI_666_924_selected-regions.csv', 1260461951),
                               ('VI_777_234/VI_777_234.eaf', 1824613251),
                               ('VI_777_234/VI_777_234.pfsx', 3383994712),
                               ('VI_777_234/VI_777_234_selected-regions.csv', 1815748137)]

    def check_first_run_outputs():
        for relative_path, checksum in expected_file_checksums:
            assert text_file_checksum(tmp_path / relative_path) == checksum

    check_first_run_outputs()

    # Make a list with one recording already processed and one new one
    info_spreadsheet_path_2 = tmp_path / 'info_spreadsheet.csv'
    info_spreadsheet_2 = pd.DataFrame(columns='id,age,length_of_recording'.split(','),
                                      data=('VI_666_924,30,960'.split(','),
                                            'VI_888_098,17,640'.split(',')))
    info_spreadsheet_2.to_csv(info_spreadsheet_path_2, index=False)
    info_spreadsheet_2.id.apply(lambda full_recording_id: tmp_path.joinpath(full_recording_id).mkdir(exist_ok=True))

    # The new run should raise an error, not touch the files created above, and not create new files.
    # No seed this time, so that if the new files do get created, they would be different
    with pytest.raises(FileExistsError):
        batch_create_files_with_random_regions(info_spreadsheet_path_2)

    # No files for the new recording
    assert not any(tmp_path.joinpath('VI_888_098').iterdir())

    # The first-run outputs have not changed
    check_first_run_outputs()


def _read_sub_recordings():
    return pd.read_csv(f'{DATA_PATH}/sub_recordings.csv',
                       parse_dates=['recording_start', 'recording_end'],
                       dtype=dict(recordings_start_wav=int))


def _read_intervals():
    return pd.read_csv(f'{DATA_PATH}/intervals.csv',
                       parse_dates=['code_onset', 'code_offset', 'context_onset', 'context_offset'],
                       dtype=dict(code_onset_wav=int))


def test_make_intervals():
    sub_recordings = _read_sub_recordings()
    actual_intervals = make_intervals(sub_recordings)
    expected_intervals = _read_intervals()

    assert_frame_equal(expected_intervals, actual_intervals)


def _read_intervals_with_fake_metric():
    return pd.read_csv(f'{DATA_PATH}/intervals_with_fake_metric.csv',
                       parse_dates=['code_onset', 'code_offset', 'context_onset', 'context_offset'],
                       dtype=dict(code_onset_wav=int))


def _read_intervals_with_vtc_total_speech_duration():
    return pd.read_csv(f'{DATA_PATH}/intervals_with_vtc_total_speech_duration.csv',
                       parse_dates=['code_onset', 'code_offset', 'context_onset', 'context_offset'],
                       dtype=dict(code_onset_wav=int))


def _read_test_vtc_data(*args, **kwargs):
    return read_rttm(Path(f'{DATA_PATH}/test_all.rttm'))


def test_add_metric():
    intervals = _read_intervals()
    vtc_data = _read_test_vtc_data()
    actual_intervals_with_metric = add_metric(intervals, vtc_data)
    expected_intervals_with_metric = _read_intervals_with_vtc_total_speech_duration()
    assert_frame_equal(actual_intervals_with_metric, expected_intervals_with_metric)


def _read_best_intervals(version: int):
    return pd.read_csv(f'{DATA_PATH}/best_intervals_{version:02}.csv',
                       dtype=dict(code_onset_wav=int, code_offset_wav=int,
                                  context_onset_wav=int, context_offset_wav=int))


def test_select_best_intervals(monkeypatch):
    # We'll only be sampling 3 intervals here
    monkeypatch.setattr(intervals_module, 'INTERVALS_FOR_ANNOTATION_COUNT', 3)

    intervals_with_metric = _read_intervals_with_fake_metric()

    # Test when there are no pre-existing interval
    actual_best_intervals_1 = select_best_intervals(intervals_with_metric, n_to_select=3)
    expected_best_intervals_1 = _read_best_intervals(1)
    assert_frame_equal(actual_best_intervals_1, expected_best_intervals_1)

    # And when there are some
    actual_best_intervals_2 = select_best_intervals(intervals_with_metric, n_to_select=3,
                                                    existing_code_intervals=PRE_EXISTING_CODE_INTERVALS)
    expected_best_intervals_2 = _read_best_intervals(2)
    assert_frame_equal(actual_best_intervals_2, expected_best_intervals_2)


def _get_test_eaf_path(*args, **kwargs):
    return Path(f'{DATA_PATH}/test_eaf.eaf')


def _get_selected_regions_path():
    return Path(f'{DATA_PATH}/selected_regions.csv')


def _get_expected_eaf_path():
    return Path(f'{DATA_PATH}/expected.eaf')


def _get_expected_with_extra_eaf_path():
    return Path(f'{DATA_PATH}/expected-with-extra.eaf')


def test_add_annotation_intervals_to_eaf(tmpdir):
    # Load inputs
    eaf = EafPlus(_get_test_eaf_path())
    best_intervals = _read_best_intervals(2)
    best_intervals.insert(1, 'sampling_type', 'high-volubility')

    # Add once
    eaf, _ = add_annotation_intervals_to_eaf(eaf, best_intervals)

    actual_eaf_path = Path(tmpdir / 'actual.eaf')
    eaf.to_file(actual_eaf_path)
    expected_eaf_path = _get_expected_eaf_path()
    assert actual_eaf_path.read_text() == expected_eaf_path.read_text()

    # Add a second time
    with pytest.raises(AssertionError):
        add_annotation_intervals_to_eaf(eaf, best_intervals)
